using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
#if GWX_CS
using GWXDeliveryDemo;
using GwxLib_CS;
using AxGwxLib_CS;
#elif GWX_HL
using GWXDeliveryDemo;
using GwxLib_HL;
using AxGwxLib_HL;
#else
using GWXLib;
using AxGWXLib;
#endif
using System.Data.OleDb;
//using System.Data.SqlClient;

namespace DeliveryDemo
{
    public partial class Form1 : Form
    {
        private AxGWControl axGWControl1;
        private DataSet input_data;
        private GWXDelivery delivery;
        private int cancontinue = 1;

        /// <summary>
        /// All program data that is stored at exit and restored at start up
        /// </summary>
        [Serializable]
        public struct Settings
        {
            public string lastmap;
            public string lastdata;
            public GWX_OptimizeMode optimize;
            public GWX_DistributionMethod distr;
            public bool multitrip;
            public bool fromgarage;
            public bool garagebetween;
            public bool togarage;
            public bool pointonce;
            public bool stacktype;
            public double intertrip;
            public bool groupdelivery;
            public bool usedate;
            public DateTime calcdate;
            public double addtime;
            public double addpath;
            public bool includetoroute;
            public bool multiday;
            public bool usetollroads;
        };

        private Settings settings;

        public struct GeoPoint
        {
            public double Lat;
            public double Lon;
        };

        public Form1()
        {
            InitializeComponent();
            settings.lastmap = "";
            settings.lastdata = "";
            settings.optimize = GWX_OptimizeMode.GWX_OptimizeTime;
            settings.distr = GWX_DistributionMethod.GWX_FavorableCombinations;
            settings.multitrip = false;
            settings.fromgarage = true;
            settings.togarage = true;
            settings.pointonce = true;
            settings.stacktype = true;
            settings.intertrip = 0;
            settings.groupdelivery = false;
            settings.addtime = 0;
            settings.addpath = 0;
            settings.includetoroute = false;
            settings.multiday = false;
            settings.usetollroads = true;
            SerializeSettings(false);
            InitButtons();
            InitGwxControl();
        }

        private void InitGwxControl()
        {
            MapControlHost.Location = new System.Drawing.Point(0, 0);
            MapControlHost.Size = splitContainer1.Panel2.ClientSize;
            axGWControl1 = new AxGWControl();
            axGWControl1.BeginInit();
            axGWControl1.Dock = DockStyle.Fill;
            axGWControl1.Enabled = true;
            axGWControl1.Location = new Point(0, 0);
            axGWControl1.Margin = new Padding(4, 4, 4, 4);
            axGWControl1.Name = "axGWControl1";
            var resources = new ComponentResourceManager(typeof (Form1));
            //axGWControl1.OcxState = ((AxHost.State)(resources.GetObject("axGWControl1.OcxState")));
            //axGWControl1.Size = new Size(736, 609);
            //axGWControl1.TabIndex = 0;
            axGWControl1.MouseAction += axGWControl1_MouseAction;
            axGWControl1.RouteProgress += axGWControl1_RouteProgress;
            MapControlHost.Controls.Add(axGWControl1);
            axGWControl1.EndInit();
        }

        private void InitButtons()
        {
#if GWX_CS
            button_load_remote_coverage.Visible = false;
#elif GWX_HL
            button_load_map.Visible = false;
            LoadStrBtn.Visible = false;
#endif
        }

        private void SerializeSettings(bool serialize)
        {
            string filename = Application.ExecutablePath; // "*.exe"
            filename = filename.Substring(0, filename.Length - 3) + "settings"; // "*.exe" -> "*.settings"

            try
            {
                XmlSerializer settingsSerializer = new XmlSerializer(typeof (Settings));
                if (serialize)
                {
// saving
                    StreamWriter sw = new StreamWriter(filename);
                    settingsSerializer.Serialize(sw, settings);
                    sw.Close();
                }
                else
                {
// restoring
                    FileStream fs = new FileStream(filename, FileMode.Open);
                    settings = (Settings) settingsSerializer.Deserialize(fs);
                    fs.Close();
                }
            }
            catch (FileNotFoundException)
            {
                // The first start or file "DeliveryDemo.settings" deleted.
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            SerializeSettings(true);
        }


        /// <summary>
        /// loads map to GWControl1 showing hourglass cursor
        /// then shows map name on status
        /// </summary>
        /// <param name="mapname"></param>
        private void loadMap(string mapname)
        {
            WaitCursor wc = new WaitCursor();
            string st;
            
#if GWX_HL
            axGWControl1.AddMap(mapname, "");
            delivery = (GWXDelivery)axGWControl1.CreateGWXDelivery("~");
            st = "Gws coverage" + (delivery == null ? ", no delivery" : ", delivery OK");
#else
            axGWControl1.MapName = mapname;
            st = axGWControl1.MapName;
            int pos = st.LastIndexOfAny(@"\/:".ToCharArray());
            if (pos >= 0) st = st.Remove(0, pos + 1);
            delivery = (GWXDelivery)axGWControl1.CreateGWXDelivery(mapname);
            st += delivery == null ? ", no delivery" : ", delivery OK";
#endif

            SetDeliveryPropertiesControls();
            toolStripStatusLabel1.Text = st;
            ShowDataOnMap();
        }

        /// <summary>
        /// slipts full file name in path and filename
        /// </summary>
        /// <param name="file">original filename</param>
        /// <param name="path">true to get path, false to get filename</param>
        /// <returns></returns>
        private static string DivideFilePath(string file, bool path)
        {
            for (int i = file.Length - 1; i >= 0; i--)
            {
                if (file[i] == '\\' || file[i] == '/' || file[i] == ':')
                    return path ? file.Substring(0, i) : file.Substring(i + 1);
            }
            return file;
        }

        /// <summary>
        /// returns path (path of DeliveryDemo.exe)\..\Data
        /// </summary>
        /// <returns></returns>
        private static string GetDefaultDataDirectory()
        {
            return DivideFilePath(DivideFilePath(Application.ExecutablePath, true), true) + @"\Data";
        }

        /// <summary>
        /// opens file dialog to get filename with specified filter
        /// </summary>
        /// <param name="filename">previous file name</param>
        /// <param name="filter">file filter i.e. "text files|*.txt"</param>
        /// <returns></returns>
        private static bool DoFileDialog(ref string filename, string filter)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = filter;
            if (filename != null && filename != "")
            {
                dlg.FileName = DivideFilePath(filename, false);
                dlg.InitialDirectory = DivideFilePath(filename, true);
            }
            else
                dlg.InitialDirectory = GetDefaultDataDirectory();

            if (dlg.ShowDialog() != DialogResult.OK)
                return false;
            filename = dlg.FileName;
            return true;
        }

        private void button_load_map_Click(object sender, EventArgs e)
        {
            if (DoFileDialog(ref settings.lastmap, "chart files|*.chart"))
                loadMap(settings.lastmap);
        }

        private void button_last_map_Click(object sender, EventArgs e)
        {
            if (settings.lastmap == null || settings.lastmap == "")
            {
                MessageBox.Show("Last loaded map is not defined. You must select chart to load.");
                button_load_map_Click(sender, e);
                return;
            }
            loadMap(settings.lastmap);
        }

        /// <summary>
        /// loads data from xml file (cars and jobs) 
        /// </summary>
        /// <param name="ask"></param>
        private void loadData(bool ask)
        {
            string filename = settings.lastdata;
            if (ask || filename == null || filename == "")
            {
                if (!DoFileDialog(ref filename, "xml and mdb files|*.xml;*.mdb"))
                    return;
            }
            WaitCursor wc = new WaitCursor();

            if (filename.ToLower().EndsWith(".mdb"))
            {
                try
                {
                    DataSet export_data = new DataSet(); // creating new

                    string ConnStr = @"Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" + filename + @";Persist Security Info=False";
                    OleDbConnection MyConn = new OleDbConnection(ConnStr);
                    MyConn.Open();
                    OleDbCommand Cmd = new OleDbCommand("select * from CARS", MyConn);
                    OleDbDataReader ObjReader = Cmd.ExecuteReader();

                    OleDbDataAdapter adapterCars = new OleDbDataAdapter("select * from CARS", MyConn);
                    adapterCars.Fill(export_data, "Cars");
                    OleDbDataAdapter adapterOrders = new OleDbDataAdapter("select * from ORDERS", MyConn);
                    adapterOrders.Fill(export_data, "Orders");
                    OleDbDataAdapter adapterZones = new OleDbDataAdapter("select * from ZONES", MyConn);
                    adapterZones.Fill(export_data, "Zones");

                    // where to save xml data?

                    SaveFileDialog dlg = new SaveFileDialog();
                    dlg.Filter = "xml files|*.xml";
                    dlg.InitialDirectory = GetDefaultDataDirectory();
                    dlg.DefaultExt = "xml";
                    dlg.Title = "Select a new xml file name to save the data";
                    //dlg.CheckPathExists = true;

                    if (dlg.ShowDialog() != DialogResult.OK)
                    {
                        input_data = export_data;
                        return;
                    }
                    string xmlfilename = dlg.FileName;

                    export_data.WriteXml(xmlfilename, XmlWriteMode.WriteSchema);

                    MyConn.Close();
                    filename = xmlfilename;
                }
                catch (Exception e)
                {
                    MessageBox.Show("Error exporting data from '" + filename + "':\r\n\r\n" + e.Message);
                    return;
                }
            }

            settings.lastdata = filename;

            input_data = new DataSet(); // creating new
            try
            {
                input_data.ReadXml(settings.lastdata, XmlReadMode.ReadSchema);
                dataGridViewCars.DataSource = input_data.Tables["Cars"];
                dataGridJobs.DataSource = input_data.Tables["Orders"];
                dataGridViewZones.DataSource = input_data.Tables["Zones"];
            }
            catch (Exception exc)
            {
                MessageBox.Show("File '" + settings.lastdata + "':\r\n\r\n" + exc.Message);
                input_data = null; // data object is invalid
            }
            ShowDataOnMap();
        }

        private void button_load_data_Click(object sender, EventArgs e)
        {
            loadData(true);
        }

        /// <summary>
        /// shows if parameters are set programmically, not by user
        /// </summary>
        private bool filling_settings = false;

        /// <summary>
        /// fills variants and sets controls from settings
        /// </summary>
        private void SetDeliveryPropertiesControls()
        {
            lb_variants.Items.Clear();

            if (delivery == null)
                return;

            GWX_VariantSubject[] subjs = new GWX_VariantSubject[]
            {
                GWX_VariantSubject.GWX_VariantWidth,
                GWX_VariantSubject.GWX_VariantHeight, GWX_VariantSubject.GWX_VariantWeight,
                GWX_VariantSubject.GWX_VariantPriority,
                GWX_VariantSubject.GWX_VariantCargo, GWX_VariantSubject.GWX_VariantLicence
            };

            foreach (GWX_VariantSubject subj in subjs)
            {
                lb_variants.Items.Add("   " + subj.ToString() + "   ");
                GWStringList list = delivery.GetVariants(subj);
                for (int i = 0; i < list.ItemCount; i++)
                    lb_variants.Items.Add(list.GetItem(i));
            }

            // settings -> controls
            filling_settings = true;
            cb_optimization.SelectedIndex = (int) settings.optimize;
            cb_distribution.SelectedIndex = (int) settings.distr;
            xb_multitrip.Checked = settings.multitrip;
            xb_fromgarage.Checked = settings.fromgarage;
            xb_garagebetween.Checked = settings.garagebetween;
            xb_togarage.Checked = settings.togarage;
            xb_pointonce.Checked = settings.pointonce;
            xb_groupdelivery.Checked = settings.groupdelivery;
            xb_stack.Checked = settings.stacktype;
            tb_interval.Text = settings.intertrip.ToString();
            tb_interval.Modified = false;
            xb_date.Checked = settings.usedate;
            dtp_date.Enabled = settings.usedate;
            dtp_date.Value = settings.usedate? settings.calcdate : DateTime.Today;
            tb_addtime.Text = settings.addtime.ToString();
            tb_addtime.Modified = false;
            tb_addpath.Text = settings.addpath.ToString();
            tb_addpath.Modified = false;
            xb_includeadditionals.Checked = settings.includetoroute;
            xb_multiday.Checked = settings.multiday;
            xb_usetollroads.Checked = settings.usetollroads;
            filling_settings = false;
        }

        /// <summary>
        /// user changed some delivery parameter
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void on_delivery_setting_change(object sender, EventArgs e)
        {
            if (filling_settings)
                return;
            if (sender == cb_optimization) settings.optimize = (GWX_OptimizeMode)cb_optimization.SelectedIndex;
            else if (sender == cb_distribution) settings.distr = (GWX_DistributionMethod)cb_distribution.SelectedIndex;
            else if (sender == xb_multitrip) settings.multitrip = xb_multitrip.Checked;
            else if (sender == xb_fromgarage) settings.fromgarage = xb_fromgarage.Checked;
            else if (sender == xb_garagebetween) settings.garagebetween = xb_garagebetween.Checked;
            else if (sender == xb_togarage) settings.togarage = xb_togarage.Checked;
            else if (sender == xb_pointonce) settings.pointonce = xb_pointonce.Checked;
            else if (sender == xb_stack) settings.stacktype = xb_stack.Checked;
            else if (sender == xb_groupdelivery) settings.groupdelivery = xb_groupdelivery.Checked;
            else if (sender == xb_includeadditionals) settings.includetoroute = xb_includeadditionals.Checked;
            else if (sender == xb_multiday) settings.multiday = xb_multiday.Checked;
            else if (sender == xb_usetollroads) settings.usetollroads = xb_usetollroads.Checked;
            else if (sender == xb_date)
            {
                settings.usedate = xb_date.Checked;
                dtp_date.Enabled = settings.usedate;
            }
            else if (sender == dtp_date) settings.calcdate = dtp_date.Value;
        }

        /// <summary>
        /// when tb_interval loses focus and is modified by user, applies new value
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tb_interval_Leave(object sender, EventArgs e)
        {
            if (!tb_interval.Modified)
                return;
            try
            {
                settings.intertrip = Convert.ToDouble(tb_interval.Text);
            }
            catch (Exception)
            {
            }
            tb_interval.Text = settings.intertrip.ToString();
            tb_interval.Modified = false;
        }

        private void tb_addtime_Leave(object sender, EventArgs e)
        {
            if (!tb_addtime.Modified)
                return;
            try
            {
                settings.addtime= Convert.ToDouble(tb_addtime.Text);
            }
            catch (Exception)
            {
            }
            tb_addtime.Text = settings.addtime.ToString();
            tb_addtime.Modified = false;
        }

        private void tb_addpath_Leave(object sender, EventArgs e)
        {
            if (!tb_addpath.Modified)
                return;
            try
            {
                settings.addpath = Convert.ToDouble(tb_addpath.Text);
            }
            catch (Exception)
            {
            }
            tb_addpath.Text = settings.addpath.ToString();
            tb_addpath.Modified = false;
        }

        private void button_last_data_Click(object sender, EventArgs e)
        {
            loadData(false);
        }

        /// <summary>
        /// shows on map cars and jobs from input_data
        /// </summary>
        private void ShowDataOnMap()
        {
            if (input_data == null || axGWControl1.MapName == "")
                return;

            axGWControl1.DeleteDBF("cars");
            axGWControl1.DeleteDBF("jobs");
            axGWControl1.DeleteDBF("joblines");
            axGWControl1.DeleteDBF("zones");

            axGWControl1.Table2Map("name='cars';descr='Cars';structure='name text Car name';", "m @502 'gwp.otl' 69",
                                   null);
            axGWControl1.Table2Map("name='jobs';descr='Jobs';structure='name text Job name, img integer image';",
                                   "m @500 'gwp.otl' [img]+65", null);
            axGWControl1.Table2Map("name='joblines';descr='Jobs';structure='name text Job name';", "l @501 red 1", null);

            axGWControl1.Table2Map("name='zones';descr='Zones';structure='id text Zone id, descr text Description, color integer Zone color';",
                "a @501 [color] 35; p @502 [color] 2", //"l @501 red 3",
                null);

            ShowCars(new MyReader(new DataTableReader(input_data.Tables["Cars"])));
            ShowJobs(new MyReader(new DataTableReader(input_data.Tables["Orders"])));
            ShowZones(new MyReader(new DataTableReader(input_data.Tables["Zones"])));
            axGWControl1.Refresh();
        }

        /// <summary>
        /// Shows cars on map
        /// </summary>
        /// <param name="mCars"></param>
        private void ShowCars(MyReader mCars)
        {
            while (mCars.Read())
            {
                double lon = mCars.GetDouble("XCOORD"), lat = mCars.GetDouble("YCOORD");
                string m = "P E " + lon.ToString("g") + " " + lat.ToString("g") + ";";
                    // metrics requires (X Y), i.e. (Lon Lat)
                axGWControl1.ModifyTable(
                    "insert into [cars] set [name]='" + mCars.GetString("N") + "', metrics='" + m + "'", 0);
            }
        }

        /// <summary>
        /// Shows jobs on map
        /// </summary>
        /// <param name="mJobs"></param>
        private void ShowJobs(MyReader mJobs)
        {
            while (mJobs.Read())
            {
                string jobName = mJobs.GetString("N");
                double lat0 = mJobs.GetDouble("GETYCOORD"), lon0 = mJobs.GetDouble("GETXCOORD");
                double lat1 = mJobs.GetDouble("PUTYCOORD"), lon1 = mJobs.GetDouble("PUTXCOORD");
                string pos0 = lon0.ToString("g") + " " + lat0.ToString("g");
                string pos1 = lon1.ToString("g") + " " + lat1.ToString("g");

                axGWControl1.ModifyTable(
                    "insert into [jobs] set [name]='" + jobName + "', [img]=0, metrics='P E " + pos0 + ";'", 0);
                axGWControl1.ModifyTable(
                    "insert into [jobs] set [name]='" + jobName + "', [img]=1, metrics='P E " + pos1 + ";'", 0);
                axGWControl1.ModifyTable(
                    "insert into [joblines] set [name]='" + jobName + "', metrics='L E " + pos0 + ", " + pos1 + ";'", 0);
            }
        }

        /// <summary>
        /// Shows zones on map
        /// </summary>
        /// <param name="mJobs"></param>
        private void ShowZones(MyReader mZones)
        {// "name='zones';descr='Zones';structure='ID text Zone id, descr text Description';"
            while (mZones.Read())
            {
                string id = mZones.GetString("ID");
                string descr = mZones.GetString("DESCR");
                string metrics = mZones.GetString("METRICS");
                int color = mZones.GetInt32("COLOR");
                if (metrics != "")
                    axGWControl1.ModifyTable(
                        "insert into [zones] set [id]='" + id + "', [descr]='" + descr + "', color=" + color.ToString() +
                        ", metrics='" + metrics + "'", 0);
            }
        }

        private void button_calc_Click(object sender, EventArgs e)
        {
            Calculate();
        }

        /// <summary>
        /// loads cars and jobs into delivery, makes delivery to calculate routes, gets results
        /// </summary>
        private void Calculate()
        {
            try
            {
                if (delivery == null)
                    throw new Exception("Delivery not created");
                if (input_data == null)
                    throw new Exception("Data not loaded");
                if (delivery.State == GWX_DeliveryState.GWX_Calculation)
                    throw new Exception("Delivery is in Calculation state");
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
                return;
            }

            delivery.Reset();

            button_abort.Enabled = true;
            button_calc.Enabled = false;
            cancontinue = 1;

            // send data to delivery
            SetDeliveryProperties();
            AddCarsToDelivery();
            AddJobsToDelivery();
            AddZonesToDelivery();

            delivery.CalculateRoutes();

            button_calc.Enabled = true;
            button_abort.Enabled = false;
            progress_calc.Value = 0;

            // filling error list
            GWStringList errors = delivery.GetErrors();
            string err_text = "";
            for (int pos = errors.moveFirst(); pos != 0; pos = errors.moveNext())
                err_text += errors.Item + "\r\n";
            tb_errors.Text = err_text;

            // show total result on status bar
            int tot_cnt = delivery.TotalJobsCount;
            int not_cnt = delivery.NotResolvedJobsCount;
            double tot_len = delivery.TotalLength/1000;
            double tot_time = delivery.TotalTime;

            string res_txt = "Length=" + tot_len.ToString("0.#") + " km, Time=" + MinutesToStrring(tot_time) +
                             "', Jobs: " + tot_cnt.ToString() + ", ";
            if (not_cnt == 0) res_txt += "all resolved";
            else res_txt += not_cnt.ToString() + " not resolved";
            toolStripStatusLabel2.Text = res_txt;
            FillCarsList(true);
            list_cars.Items[0].Selected = true;
        }

        /// <summary>
        /// set current settings to delivery
        /// </summary>
        private void SetDeliveryProperties()
        {
            delivery.ProgressDelay = 5000;
            delivery.OptimizeMode = settings.optimize;
            delivery.DistributionMethod = settings.distr;
            delivery.MultiTrip = settings.multitrip;
            delivery.PathFromGarage = settings.fromgarage;
            delivery.GarageBetweenTrip = settings.garagebetween;
            delivery.PathToGarage = settings.togarage;
            delivery.PointOnce = settings.pointonce;
            delivery.StackUnloading = settings.stacktype;
            delivery.TripInterval = settings.intertrip;
            delivery.GroupDelivery = settings.groupdelivery;
            delivery.AdditionalTimeInPoint = settings.addtime;
            delivery.AdditionalDistanceInPoint = settings.addpath;
            delivery.IncludeAdditionsInRoute = settings.includetoroute;
            delivery.MutliDayRoutes = settings.multiday;
            delivery.UseTollRoads = settings.usetollroads? 1 : 0;
        }

        /// <summary>
        /// called when calculations are in progress, attached to axGWControl1 as event handler
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void axGWControl1_RouteProgress(object sender, _IGWControlEvents_RouteProgressEvent e)
        {
            progress_calc.Value = e.percentDone;
            e.canContinue = cancontinue;
            Application.DoEvents();
        }

        //private void axGWControl1_AsyncReqComplete(object sender, AxGWXLib._IGWControlEvents_AsyncReqCompleteEvent e)
        //{
        //    String s = e.response;
        //}

        private void button_abort_Click(object sender, EventArgs e)
        {
            cancontinue = 0;
        }

        private DateTime ZeroDate = DateTime.MinValue;
        private DateTime EmptyDate = new DateTime(1899,12,30);
        private DateTime CheckDefaultDate(DateTime dt, DateTime dt_prev)
        { // dt = {30.12.1899 1:05:00}
            if(!settings.usedate)
                return dt;
            if (dt.Date != EmptyDate.Date)
                return dt;
            dt = settings.calcdate.Date + dt.TimeOfDay;
            if (dt_prev > dt)
                dt.AddDays(1);
            return dt;
        }
        /// <summary>
        /// fills car table in delivery from input_data
        /// </summary>
        private void AddCarsToDelivery()
        {
            MyReader mCars = new MyReader(new DataTableReader(input_data.Tables["Cars"]));
            while (mCars.Read())
            {
                GWXDeliveryCar car = delivery.AddCar();

                car.Name = mCars.GetString("N");
                car.Description = mCars.GetString("DESCR");

                car.MaxWeight = mCars.GetDouble("MAXWEIGHT");
                car.MaxVolume = mCars.GetDouble("MAXVOLUME");
                car.MaxVolume2 = mCars.GetDouble("MAXVOLUME2");
                car.MaxCost = mCars.GetDouble("MAXCOST");
                car.MaxJobNumber = mCars.GetInt32("MAXORD");
                car.MaxRouteLength = mCars.GetDouble("MAXLEN");
                car.MaxRouteTime = mCars.GetDouble("MAXTIME")*60.0; // hours -> mins

                car.OwnWeight = mCars.GetDouble("SELFWEIGHT");
                car.SetGarage(mCars.GetDouble("YCOORD"), mCars.GetDouble("XCOORD"), mCars.GetString("GARAGE"));

                car.SpeedCoefficient = mCars.GetDouble("SPEEDK", 1);
                DateTime dt1, dt2;
                if (mCars.GetDateTime("WRKBEG", out dt1)) car.WorkBegin = CheckDefaultDate(dt1, ZeroDate);
                if (mCars.GetDateTime("WRKEND", out dt2)) car.WorkEnd = CheckDefaultDate(dt2, dt1);

                if (mCars.GetDateTime("LUNCHBEG", out dt1)) car.LunchBegin = dt1; // date not used
                if (mCars.GetDateTime("LUNCHEND", out dt1)) car.LunchEnd = dt1;

                car.Width = mCars.GetDouble("WIDTH");
                car.Height = mCars.GetDouble("HEIGHT");
                car.Cargo = mCars.GetDouble("CARGO");
                car.Licence = mCars.GetString("LICENCE");

                car.CarType = mCars.GetString("CARTYPE");
                car.FuelConsumption = mCars.GetDouble("FUELPERKM");

                car.LoadZoneList = mCars.GetString("LOADZONES");
                car.UnloadZoneList = mCars.GetString("UNLOADZONES");
                car.WorkTime = mCars.GetString("WORKTIME");
                car.UseTollRoads = mCars.GetInt32("USETOLLROARDS", -1);
            }
        }

        /// <summary>
        /// fills job table in delivery from input_data
        /// </summary>
        private void AddJobsToDelivery()
        {
            MyReader mJobs = new MyReader(new DataTableReader(input_data.Tables["Orders"]));
            while (mJobs.Read())
            {
                GWXDeliveryJob job = delivery.AddJob();

                job.Name = mJobs.GetString("N");
                job.Priority = mJobs.GetInt32("PRIORITY");

                job.Weight = mJobs.GetDouble("WEIGHT");
                job.Volume = mJobs.GetDouble("VOLUME");
                job.VolumeDisp = mJobs.GetInt32("VOLUME_DISP");
                job.Cost = mJobs.GetDouble("COST");
                job.CarTypes = mJobs.GetString("CARTYPES");
                job.MaxTransportDuration = mJobs.GetDouble("MAXTRANSPTIME");
                    // already in minutes *60.0; // hours -> mins

                DateTime dt1, dt2;
                job.SetGetPoint(mJobs.GetDouble("GETYCOORD"), mJobs.GetDouble("GETXCOORD"), mJobs.GetString("GETDESCR"));
                if (mJobs.GetDateTime("GETBEG", out dt1)) job.GetTimeAfter = CheckDefaultDate(dt1, ZeroDate);
                if (mJobs.GetDateTime("GETEND", out dt2)) job.GetTimeBefore = CheckDefaultDate(dt2, dt1);
                job.GetDuration = mJobs.GetDouble("GETDUR");

                job.SetPutPoint(mJobs.GetDouble("PUTYCOORD"), mJobs.GetDouble("PUTXCOORD"), mJobs.GetString("PUTDESCR"));
                if (mJobs.GetDateTime("PUTBEG", out dt1)) job.PutTimeAfter = CheckDefaultDate(dt1, ZeroDate);
                if (mJobs.GetDateTime("PUTEND", out dt2)) job.PutTimeBefore = CheckDefaultDate(dt2, dt1);
                job.PutDuration = mJobs.GetDouble("PUTDUR");

                job.LoadTime = mJobs.GetString("LOADTIME");
                job.UnloadTime = mJobs.GetString("UNLOADTIME");


                job.CarName = mJobs.GetString("CARN");
                job.IndexInRouteGet = mJobs.GetInt32("RTGETIX");
                job.IndexInRoutePut = mJobs.GetInt32("RTPUTIX");
            }
        }

        /// <summary>
        /// fills zones table in delivery from input_data
        /// </summary>
        private void AddZonesToDelivery()
        {
            MyReader mZones = new MyReader(new DataTableReader(input_data.Tables["Zones"]));
            while (mZones.Read())
            {
               // GWXDeliveryJob job = delivery.AddJob();
                string id = mZones.GetString("ID");
                string metrics = mZones.GetString("METRICS");
                delivery.AddZone(id, metrics);
            }
        }

        private IEnumerable<GWXDeliveryCar> CarIterator()
        {
            GWXDeliveryCar car;
            for (int i = 0; (car = delivery.GetCar(i)) != null; i++)
                yield return car;
        }

        private IEnumerable<GWXDeliveryJob> JobIterator()
        {
            GWXDeliveryJob job;
            for (int i = 0; (job = delivery.GetJob(i)) != null; i++)
                yield return job;
        }

        /// <summary>
        /// converts to string in format 26 -> "26", 156.5 -> "2:37"
        /// </summary>
        /// <param name="minutes"></param>
        /// <returns></returns>
        private static string MinutesToStrring(double minutes)
        {
            int m = (int) (minutes + 0.5);
            if (m < 60) return m.ToString();
            int h = m/60;
            m = m%60;
            return h.ToString() + ":" + m.ToString("00");
        }

        /// <summary>
        /// adds or updates item in car list
        /// </summary>
        /// <param name="firsttime">add new item</param>
        /// <param name="i">row index</param>
        /// <param name="name">car name</param>
        /// <param name="jobsCount">jobs count</param>
        /// <param name="length">route length</param>
        /// <param name="time">route time</param>
        /// <param name="errCount">count of errors</param>
        private ListViewItem AddToCarList(bool firsttime, ref int i, string name, int jobsCount, double length, double time,
                                  int errCount)
        {
            ListViewItem lwi = null;
            if (firsttime)
            {
                lwi = new ListViewItem(name);
                list_cars.Items.Add(lwi);
                for (int k = 0; k < 4; k++)
                    lwi.SubItems.Add("");
            }
            else
            {
                lwi = list_cars.Items[i++];
            }
            // name is the same
            lwi.SubItems[1].Text = jobsCount.ToString();
            lwi.SubItems[2].Text = length > 0.0 ? length.ToString("0.#") : "";
            lwi.SubItems[3].Text = time > 0.0 ? MinutesToStrring(time) : "";
            lwi.SubItems[4].Text = errCount > 0 ? errCount.ToString() : "";
            return lwi;
        }

        /// <summary>
        /// Fills result car list + items "all" and "not resolved"
        /// </summary>
        /// <param name="firsttime">true to add items for the first time, false for update existing items</param>
        private void FillCarsList(bool firsttime) // firsttime - fill list, else update items
        {
            if (firsttime)
                list_cars.Items.Clear();
            int tot_cnt = delivery.TotalJobsCount;
            int not_cnt = delivery.NotResolvedJobsCount;
            double tot_len = delivery.TotalLength/1000;
            double tot_time = delivery.TotalTime;
            int ii = 0;
            // summary
            AddToCarList(firsttime, ref ii, "<  all  >", tot_cnt, tot_len, tot_time, 0);
            int errCnt = 0;
            foreach (GWXDeliveryCar car in CarIterator())
            {
                AddToCarList(firsttime, ref ii, car.Name, car.ResJobCount, car.ResRouteLength, car.ResRouteTime,
                             car.ResViolationsCount);
                errCnt += car.ResViolationsCount;
            }
            // not distributed
            ListViewItem lwi = AddToCarList(firsttime, ref ii, "<not resolved>", not_cnt, 0.0, 0.0, 0);
            if (errCnt > 0)
                lwi.SubItems[4].Text = errCnt.ToString();
        }

        private const string ROUTETABLEHEADER = "Route";
        private const string POINTSHEADER = "Points of car ";

        /// <summary>
        /// Removes all lines of cars from map
        /// </summary>
        private void RemoveCarLines()
        {
            GWStringList list = axGWControl1.DBFLoadedList;
            for (int pos = list.moveFirst(); pos != 0; pos = list.moveNext())
            {
                string db = list.Item;
                if (db.StartsWith(ROUTETABLEHEADER) || db.StartsWith(POINTSHEADER) || db == "joblines")
                    axGWControl1.DeleteDBF(db);
            }
        }

        /// <summary>
        /// shows lines of route for specified car
        /// </summary>
        /// <param name="car"></param>
        private void ShowCarLines(GWXDeliveryCar car, bool point_times)
        {
            string name = car.Name;
            string loadcmd = "name='" + ROUTETABLEHEADER + name + "';descr='" + ROUTETABLEHEADER + name +
                             "';metrics=[Metrics];";
            string style = "p @500 Crimson 205";
            var tbl = car.GetRoute();
            if (tbl == null)
                return;
            axGWControl1.Table2Map(loadcmd, style, tbl);

            if (point_times)
            {
                GWTable crdTbl = axGWControl1.CreateGWTable();
                crdTbl.addColumn("DOUBLE", "Lat", "Latitude");
                crdTbl.addColumn("DOUBLE", "Lon", "Lonfitude");
                crdTbl.addColumn("TEXT", "Text", "Point name and time");
                crdTbl.addColumn("INTEGER", "Marker", "Marker");
                crdTbl.addColumn("INTEGER", "Level", "Priority");

                for( int pos = tbl.moveFirst(); pos >=0; pos = tbl.moveNext() )
                {
                    int ptype = (int)tbl.getValue(0);
                    if (ptype == 7 || ptype == 6)
                    {
                        crdTbl.addNew();
                        crdTbl.setValue(tbl.getValue(1), 0); // lat
                        crdTbl.setValue(tbl.getValue(2), 1); // lon
                        crdTbl.setValue("Pause", 2); // text
                        crdTbl.setValue(ptype == 7? 77 : 75, 3); // marker
                        crdTbl.setValue(1, 4);
                    }
                    double prevLat = 0, prevLon = 0;
                    DateTime prevTime = new DateTime(0);
                    DateTime Time = new DateTime(0);
                    GWTable pathTbl = tbl.getTable(12); // path from the previous point to this
                    for( int j = pathTbl.moveFirst(); j >=0; j =pathTbl.moveNext() )
                    {// 0-lat 1-lon (double) 2-name(text), 4-time_text
                        crdTbl.addNew();
                        crdTbl.setValue( pathTbl.getValue( 0 ), 0 );
                        crdTbl.setValue( pathTbl.getValue( 1 ), 1 );
                        crdTbl.setValue( pathTbl.getText( 4 ) + " - " + pathTbl.getText( 2 ), 2 );
                        crdTbl.setValue(0, 4);
                        double Lat = (double)pathTbl.getValue(0);
                        double Lon = (double) pathTbl.getValue(1);
                        try
                        {
                            var vt = pathTbl.getValue(3);
                            Time = (DateTime)vt;
                        }
                        catch (Exception e)
                        {
                            string timestr = pathTbl.getText(4);
                            prevLat = prevLon = 0;
                        }
                        int wait = 0;
                        if (prevLat == Lat && prevLon == Lon)
                        {
                            TimeSpan ts = Time - prevTime;
                            wait = (int)ts.TotalSeconds;
                            if (wait < 0) // if midnight passed
                                wait += 60 * 60 * 24;
                        }
                        crdTbl.setValue(wait >= 60? 22 : 6, 3); // wait more then 1 min -> '[P]' marker instead of 'o'
                        prevLat = Lat;
                        prevLon = Lon;
                        prevTime = Time;
                    }
                }
                string loadcmd2 = "name='" + POINTSHEADER + name + "';descr='" + POINTSHEADER + name + "';coords=[Lat],[Lon];";
                axGWControl1.Table2Map(loadcmd2, "if([Level]>0) { m @503 'gwp.otl' [Marker] } else { m @502 'gwp.otl' [Marker] }", crdTbl);
            }
        }

        /// <summary>
        /// adds column headers with specified text and width
        /// </summary>
        /// <param name="colheaders">like "Name:90;Car:90;Weight:50;"</param>
        private void PrepareJobGridColumns(string colheaders) // "Name:90;Car:90;Weight:50;"
        {
            DataGridViewCell cell = new DataGridViewTextBoxCell();
            foreach (string coldef in colheaders.Split(';'))
            {
                int sep = coldef.IndexOf(':');
                if (sep < 0) continue;
                DataGridViewColumn col = new DataGridViewColumn();
                col.HeaderText = coldef.Substring(0, sep);
                col.Width = Convert.ToInt32(coldef.Substring(sep + 1));
                col.CellTemplate = cell;
                grid_jobs.Columns.Add(col);
            }
        }

        /// <summary>
        /// means of point types
        /// </summary>
        private static string[] point_types = { "begin", "load", "unload", "end", "trip end", "not distributed", "lunch", "interval" };

        /// <summary>
        /// shows current value and its increment with + or - when delta !=0
        /// </summary>
        /// <param name="value">current value</param>
        /// <param name="delta">increment</param>
        /// <returns></returns>
        private static string ValueAndDelta(double value, double delta)
        {
            string fmt = delta > 0.0 ? "{0:0.#} (+{1:0.#})" : delta < 0.0 ? "{0:0.#} ({1:0.#})" : "{0:0.#}";
            return string.Format(fmt, value, delta);
        }

        /// <summary>
        /// means of error numbers
        /// </summary>
        private static string[] error_means =
        {
            "Error", "Point visited more than once", "Max weight is exceeded by {0}", "Max volume is exceeded by {0}",
            "Max cost is exceeded by {0}", "Max order number is exceeded by {0}", "Car type mismatch",
            "Handling time exceeded by {0}",
            "Earlier by {0}", "Later by {0}", "Max duration exceeded by {0}", "Max length exceeded by {0}",
            "Work ends later by {0}",
            "Zone condition is violated",
            "Max volume 2 is exceeded by {0}", //#14#
            "Different groups" //"#15#"
        };

        /// <summary>
        /// converts error string returned in route table
        /// </summary>
        /// <param name="errmsg">like "#nr#value#nr2#value2"</param>
        /// <returns></returns>
        private static string TranslateError(string errmsg)
        {
// #nr#value#nr2#value2 
            string res = "";
            int pos0;
            while ((pos0 = errmsg.IndexOf('#')) >= 0)
            {
                pos0++;
                int pos1 = errmsg.IndexOf('#', pos0);
                if (pos1 < pos0)
                    break;
                int nr = 0;
                try
                {
                    nr = Convert.ToInt32(errmsg.Substring(pos0, pos1 - pos0));
                }
                catch (Exception)
                {
                    break;
                }

                pos0 = pos1 + 1;
                pos1 = errmsg.IndexOf('#', pos0);
                if (pos1 < 0) pos1 = errmsg.Length;

                if (res != "") res += " / ";
                res += string.Format(error_means[nr], errmsg.Substring(pos0, pos1 - pos0));
                errmsg = errmsg.Substring(pos1);
            }
            return res;
        }

        /// <summary>
        /// in different modes to fill job list, dependiong on selected item in car list, where is column with job name
        /// </summary>

        private void list_cars_SelectedIndexChanged(object sender, EventArgs e)
        {
            FillJobList();
        }


        /// <summary>
        /// fills job list in different modes depending on what is selected in car list - car, all or 'not distributed jobs' item
        /// </summary>
        private void FillJobList()
        {
            carOfJobList = "";
            routeGeoPoints.Clear();
            ListView.SelectedIndexCollection ixs = list_cars.SelectedIndices;
            int ix = -1;
            if (ixs.Count == 1) ix = ixs[0];
            RemoveCarLines();
            grid_jobs.Rows.Clear();
            grid_jobs.Columns.Clear();
            lb_route_descr.Items.Clear();
            if (ix < 0)
                return;
            if (ix == 0)
            {
// all cars
                foreach (GWXDeliveryCar car in CarIterator())
                    ShowCarLines(car, false);

                PrepareJobGridColumns("Job:60;Car:60;GetIx  PutIx:80;Weight:50;Volume:50;Cost:45;");
                foreach (GWXDeliveryJob job in JobIterator())
                {
                    if (job.CarName != "")
                        grid_jobs.Rows.Add(job.Name, job.CarName,
                                           string.Format("{0}  {1}", job.IndexInRouteGet, job.IndexInRoutePut),
                                           job.Weight, job.Volume, job.Cost);
                }
                jobNameColumn = 0;
            }
            else if (ix == list_cars.Items.Count - 1) // list_cars.Items.Count-1 - <not resolved orders>
            {
                PrepareJobGridColumns("Job:60;Weight:50;Volume:50;Cost:45;");
                foreach (GWXDeliveryJob job in JobIterator())
                {
                    if (job.CarName == "")
                        grid_jobs.Rows.Add(job.Name, job.Weight, job.Volume, job.Cost);
                }
                jobNameColumn = 0;
            }
            else
            {
                // only one car
                GWXDeliveryCar car = delivery.GetCar(list_cars.Items[ix].Text);
                carOfJobList = car.Name;
                ShowCarLines(car, true);
                GWTable tbl = car.GetRoute();
                FillJobGrid(tbl);
                FillRouteList(tbl);
            }
        }

        private void FillJobGrid(GWTable tbl)
        {
            /* Table columns:
                0  "Type",     "INTEGER", "Point type"
                1  "Lat",      "DOUBLE",  "Latitude"
                2  "Lon",      "DOUBLE",  "Longitude"
                3  "Time",     "DOUBLE",  "Arrival time"
                4  "TimeText", "TEXT",    "Arrival time as text"
                5  "Weight",   "DOUBLE",  "Job weight"
                6  "Volume",   "DOUBLE",  "Job volume"
                7  "Cost",     "DOUBLE",  "Job cost"
                8  "Job",      "TEXT",    "Job name"
                9  "Error",    "TEXT",    "Error message"
                10 "Length",   "DOUBLE",  "Path length"
                11 "Duration", "DOUBLE",  "Path duration"
                12 "Path",     "TABLE",   "Path as table"
                13 "Metrics",  "TEXT",    "Path as metrics string"
                14 "Description", "TEXT", "Point description"
                15 "VolumeDisp",  "INTEGER", "Volume 1 or 2"
             */
            PrepareJobGridColumns(
                "Type:40;Job:60;Description:80;Time:50;Weight:70;Volume:100;Cost:50;Length:50;Time:50;Error:120");
            //                     0       8,     14             4       5         6         7       10        11      9
            double sum_weight = 0, sum_volume1 = 0, sum_volume2 = 0, sum_cost = 0;
            for (int pos = tbl.moveFirst(); pos >= 0; pos = tbl.moveNext())
            {
                int ptype = (int) tbl.getValue(0);
                double weight = (double) tbl.getValue(5),
                       volume1 = (double) tbl.getValue(6), volume2 = 0,
                       cost = (double) tbl.getValue(7);
                int volume_disp = (int)tbl.getValue(15);
                sum_weight += weight;
                if (volume_disp == 2)
                {
                    volume2 = volume1;
                    volume1 = 0;
                }
                sum_volume1 += volume1;
                sum_volume2 += volume2;
                if (ptype == 1) sum_cost += cost;
                else cost = 0;
                int gridIx = grid_jobs.Rows.Add(point_types[ptype], tbl.getText(8), tbl.getText(14), tbl.getText(4),
                                   // type, job name, point_descr, time str, 
                                   ValueAndDelta(sum_weight, weight),
                                   ValueAndDelta(sum_volume1, volume1) + " / " + ValueAndDelta(sum_volume2, volume2),
                                   ValueAndDelta(sum_cost, cost), // weight, volume, cost
                                   ((double) tbl.getValue(10)).ToString("0.#"),
                                   MinutesToStrring((double) tbl.getValue(11)), // length, time
                                   TranslateError(tbl.getText(9)) // error
                    );
					
                GeoPoint gp;
                gp.Lat = (double)tbl.getValue(1);
                gp.Lon = (double)tbl.getValue(2);
                routeGeoPoints.Add(gp);
            }
            jobNameColumn = 1;
        }

        /// <summary>
        /// fills route list with point names and road segment names
        /// </summary>
        /// <param name="tbl">path table (lat, lon, name)</param>
        private void FillRouteList(GWTable tbl)
        {
            for (int pos = tbl.moveFirst(); pos >= 0; pos = tbl.moveNext())
            {
                GWTable tblrt = tbl.getTable(12); // route from previous point
                string last_street = "";
                if (tblrt != null)
                    for (int posrt = tblrt.moveFirst(); posrt >= 0; posrt = tblrt.moveNext())
                    {
                        string street = tblrt.getText(2);
                        if (street == "") continue; // street = "<no name>";
                        if (last_street == street) continue;
                        string tm = tblrt.getText(4); // time in point
                        if (tm != "")
                            tm += " ";
                        lb_route_descr.Items.Add(tm + "  - " + street);
                        last_street = street;
                    }

                int ptype = (int) tbl.getValue(0);
                string pointdescr = tbl.getText(4);
                pointdescr += " - " + point_types[ptype];
                if (ptype == 1 || ptype == 2) // 1-load, 2-unload
                    pointdescr += " job " + tbl.getText(8);
                string p_descr = tbl.getText(14);
                if (p_descr != "") pointdescr += " [" + p_descr + "]";

                lb_route_descr.Items.Add(pointdescr);
            }
        }

        private Point grid_jobs_last_point = new Point(-1, -1);
        private int grid_jobs_dragged_index = -1;
        private int grid_jobs_drop_min_index, grid_jobs_drop_max_index;
        private string dragJobName = "";
        private string dragJobCarName = "";
        private int jobNameColumn = 0;
        private string carOfJobList = ""; // name of car for which the resulting job list is shown
        List<GeoPoint> routeGeoPoints = new List<GeoPoint>(); // accordingly to the currently shown points in job_grid for one car or empty

        /// <summary>
        /// hit test job grid
        /// </summary>
        /// <param name="x">x</param>
        /// <param name="y">y</param>
        /// <param name="is_client_point">client point or screen point</param>
        /// <returns></returns>
        private int JobIndexFromPoint(int x, int y, bool is_client_point)
        {
            if (!is_client_point)
            {
                Point p = grid_jobs.PointToClient(new Point(x, y));
                x = p.X;
                y = p.Y;
            }
            DataGridView.HitTestInfo info = grid_jobs.HitTest(x, y);
            if (info.Type == DataGridViewHitTestType.Cell) return info.RowIndex;
            if (info.Type == DataGridViewHitTestType.None && grid_jobs.Rows.Count > 0)
            {
                Rectangle rect = grid_jobs.GetRowDisplayRectangle(grid_jobs.Rows.Count - 1, false);
                if (y > rect.Bottom && x < rect.Right)
                    return grid_jobs.Rows.Count; // after the last
            }
            return -1;
        }

        /// <summary>
        /// preparing to drag'n'drop
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_jobs_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Left)
                return;
            int jobIx = JobIndexFromPoint(e.X, e.Y, true);

            grid_jobs_dragged_index = -1;
            grid_jobs_last_point.X = -1;
            grid_jobs_drop_min_index = grid_jobs_drop_max_index = 0;

            if (jobIx >= 0 && jobIx < grid_jobs.Rows.Count)
            {
                dragJobCarName = carOfJobList;
                dragJobName = grid_jobs.Rows[jobIx].Cells[jobNameColumn].Value.ToString(); // find out job name
                if (dragJobName != "")
                {
// prepare to drag
                    GWXDeliveryJob job = delivery.GetJob(dragJobName);
                    if (job != null)
                    {
//                        dragJobCarName = job.CarName;
                        grid_jobs_dragged_index = jobIx;
                        grid_jobs_last_point = e.Location;
                    }
                }
                else if (IsAuxPoint(jobIx))
                {// the dragged item is Lunch or Trip End
                    grid_jobs_dragged_index = jobIx;
                    grid_jobs_last_point = e.Location;
                }
            }
        }

        private bool IsAuxPoint(int ix)
        {
            if (jobNameColumn != 1 || ix<0 || ix>= grid_jobs.Rows.Count)
                return false;
            string ptype = grid_jobs.Rows[ix].Cells[0].Value.ToString();
            return (ptype == point_types[4] || ptype == point_types[6]);
        }

        private bool DraggingAuxPoint()
        {
            return IsAuxPoint(grid_jobs_dragged_index);
        }

        /// <summary>
        /// are two points closer then delta?
        /// </summary>
        /// <param name="p1"></param>
        /// <param name="p2"></param>
        /// <param name="delta"></param>
        /// <returns></returns>
        private static bool InDelta(Point p1, Point p2, int delta)
        {
            int dx = p1.X - p2.X, dy = p1.Y - p2.Y;
            if ((Math.Abs(dx) > delta || Math.Abs(dy) > delta))
                return false;
            return dx*dx + dy*dy <= delta*delta;
        }

        /// <summary>
        /// if mouse moved enough begins drag and drop
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_jobs_MouseMove(object sender, MouseEventArgs e)
        {
            const int delta = 3;
            Point p = e.Location;
            if (grid_jobs_dragged_index == -1 || grid_jobs_last_point.X == -1 || InDelta(grid_jobs_last_point, p, delta) ||
                e.Button != MouseButtons.Left)
                return;

            grid_jobs_last_point.X = -1;
            lastDropSource = grid_jobs;

            if (list_cars.Items[0].Selected || list_cars.Items[list_cars.Items.Count - 1].Selected)
                grid_jobs_drop_min_index = grid_jobs_drop_max_index = 0;
                    //        ,      
            else
            {
                grid_jobs_drop_min_index = grid_jobs_drop_max_index = grid_jobs_dragged_index;

                int minix = 0, maxix = grid_jobs.Rows.Count; // minimal and maximal possible drop index
                if (grid_jobs.Rows[minix].Cells[jobNameColumn].Value.ToString() == "") minix++; // begin point
                if (grid_jobs.Rows[maxix - 1].Cells[jobNameColumn].Value.ToString() == "") maxix--; // end point

                if (DraggingAuxPoint()) // we can move lunch where we want to
                {
                    grid_jobs_drop_min_index = minix;
                    grid_jobs_drop_max_index = maxix;
                }
                else
                {
                    for (int i = minix; i < maxix; i++)
                    {
                        // searching other inex of this job
                        if (i == grid_jobs_dragged_index) continue; // we need other index
                        string jname = grid_jobs.Rows[i].Cells[jobNameColumn].Value.ToString();
                        if (jname == dragJobName) // i is second index
                        {
                            if (i > grid_jobs_dragged_index) // dragging before i
                            {
                                grid_jobs_drop_min_index = minix;
                                grid_jobs_drop_max_index = i;
                            }
                            else // dragging after
                            {
                                grid_jobs_drop_min_index = i + 1;
                                grid_jobs_drop_max_index = maxix;
                            }
                            break;
                        }
                    }
                }
            }

            DragDropEffects dde = DoDragDrop(grid_jobs, DragDropEffects.Move);
            lastDropSource = null;
            grid_jobs_dragged_index = -1;
            InvalidateDropTarget();
        }

        private void grid_jobs_MouseUp(object sender, MouseEventArgs e)
        {
            grid_jobs_last_point.X = -1;
            grid_jobs_dragged_index = -1;
        }

        private void list_cars_DragEnter(object sender, DragEventArgs e)
        {
        }

        private Control lastDropTarget = null;
        private Control lastDropSource = null;
        private int lastDropTargetIndex = -1;
        private Rectangle lastDropTargetInvalidateRect = Rectangle.Empty;

        private void InvalidateDropTarget()
        {
            if (lastDropTarget != null && !lastDropTargetInvalidateRect.IsEmpty)
                lastDropTarget.Invalidate(lastDropTargetInvalidateRect);
            lastDropTarget = null;
            lastDropTargetIndex = -1;
            lastDropTargetInvalidateRect = Rectangle.Empty;
        }

        /// <summary>
        /// fills with semi-transparent red the area of drop target
        /// </summary>
        /// <param name="target"></param>
        /// <param name="index"></param>
        private void SetDropTarget(Control target, int index)
        {
            if (lastDropTarget == target && lastDropTargetIndex == index)
                return;
            InvalidateDropTarget();
            if (target == null)
                return;
            lastDropTarget = target;
            lastDropTargetIndex = index;

            if (target == list_cars)
            {
                lastDropTargetInvalidateRect = list_cars.Items[index].Bounds;
            }
            else if (target == grid_jobs)
            {
                if (index == grid_jobs.Rows.Count)
                {
                    lastDropTargetInvalidateRect = grid_jobs.GetRowDisplayRectangle(index - 1, true);
                    lastDropTargetInvalidateRect.Offset(0, lastDropTargetInvalidateRect.Height);
                }
                else lastDropTargetInvalidateRect = grid_jobs.GetRowDisplayRectangle(index, true);
                lastDropTargetInvalidateRect.Height = 8;
                lastDropTargetInvalidateRect.Offset(0, -4);
            }
            using (Graphics gr = target.CreateGraphics())
            using (Brush br = new SolidBrush(Color.FromArgb(100, Color.MediumVioletRed)))
                gr.FillRectangle(br, lastDropTargetInvalidateRect);
        }


        private void list_cars_DragLeave(object sender, EventArgs e)
        {
            SetDropTarget(null, -1);
        }

        /// <summary>
        /// hit test of car list
        /// </summary>
        /// <param name="e"></param>
        /// <returns></returns>
        private int CarIndexFromPoint(DragEventArgs e)
        {
            Point p = list_cars.PointToClient(new Point(e.X, e.Y));
            ListViewHitTestInfo hti = list_cars.HitTest(p);
            if (hti.Item == null)
                return -1;
            if (p.Y < hti.Item.Bounds.Top || p.Y >= hti.Item.Bounds.Bottom)
                return -1;
            if (hti.Item.Selected) // not for already selected car
                return -1;
            return hti.Item.Index;
        }

        private void list_cars_DragOver(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs || DraggingAuxPoint())
                return;
            int carIx = CarIndexFromPoint(e);
            if (carIx > 0 && // not -1 and not "all cars"
                dragJobCarName != list_cars.Items[carIx].Text) // not to the same car
            {
                e.Effect = DragDropEffects.Move;
                SetDropTarget(list_cars, carIx);
            }
            else
            {
                e.Effect = DragDropEffects.None;
                SetDropTarget(null, -1);
            }
        }

        private void grid_jobs_DragEnter(object sender, DragEventArgs e)
        {
        }

        private void grid_jobs_DragLeave(object sender, EventArgs e)
        {
            SetDropTarget(null, -1);
        }

        private void grid_jobs_DragOver(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs)
                return;
            // we do not move job if there are all connected jobs shown
            int jobIx = grid_jobs_drop_min_index != grid_jobs_drop_max_index ? JobIndexFromPoint(e.X, e.Y, false) : -1;

            if (jobIx >= grid_jobs_drop_min_index && jobIx <= grid_jobs_drop_max_index && // 0 is route beginning
                jobIx != grid_jobs_dragged_index && jobIx != grid_jobs_dragged_index + 1) // not adjacent
            {
                e.Effect = DragDropEffects.Move;
                SetDropTarget(grid_jobs, jobIx);
            }
            else
            {
                e.Effect = DragDropEffects.None;
                SetDropTarget(null, -1);
            }


        }

        private void list_cars_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Effect != DragDropEffects.Move)
                return;
            int carIx = CarIndexFromPoint(e);
            if (carIx <= 0) // not to "all cars"
                return;
            string targetCarName = carIx < list_cars.Items.Count - 1 ? list_cars.Items[carIx].Text : "";
            if (dragJobCarName == targetCarName) // not to the same car
                return;

            MoveDraggedJobToCar(targetCarName);
        }

        private void grid_jobs_DragDrop(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs)
                return;

            if (grid_jobs_drop_min_index == grid_jobs_drop_max_index)
                return;
            int jobIx = JobIndexFromPoint(e.X, e.Y, false);

            if (jobIx < grid_jobs_drop_min_index || jobIx > grid_jobs_drop_max_index || // overrun
                jobIx == grid_jobs_dragged_index || jobIx == grid_jobs_dragged_index + 1) // adjacent
                return;

            MoveDraggedJobToPosition(jobIx);
            grid_jobs.Rows[grid_jobs_dragged_index < jobIx ? jobIx - 1 : jobIx].Selected = true;
        }

        /// <summary>
        /// assigning new car to job (put dropped job in target car)
        /// </summary>
        /// <param name="targetCarName"></param>
        private void MoveDraggedJobToCar(string targetCarName)
        {
            WaitCursor wc = new WaitCursor();
            GWXDeliveryJob job = delivery.GetJob(dragJobName);
            job.CarName = targetCarName; // long calculations in delivery
            UpdateCarAndJobLists();
        }

        /// <summary>
        /// assigning new point index in route in the same car (move dropped point in target position)
        /// </summary>
        /// <param name="jobIx"></param>
        private void MoveDraggedJobToPosition(int jobIx)
        {
            WaitCursor wc = new WaitCursor();

            if (DraggingAuxPoint()) // now do nothing if dragging lunch
            {
                GWXDeliveryCar car = delivery.GetCar(dragJobCarName);
                car.ChangeRoutePointIndex(grid_jobs_dragged_index, jobIx);
            }
            else
            {
                GWXDeliveryJob job = delivery.GetJob(dragJobName);
                if (grid_jobs_dragged_index == job.IndexInRouteGet)
                    job.IndexInRouteGet = jobIx; // long calculations in delivery
                else
                    job.IndexInRoutePut = jobIx; // long calculations in delivery
            }
            UpdateCarAndJobLists();
        }

        private void UpdateCarAndJobLists()
        {
            // updating cars list (at place)
            FillCarsList(false);
            // updateing job list
            FillJobList();
        }

        /// <summary>
        /// initializing form on start up
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form1_Shown(object sender, EventArgs e)
        {
            if (settings.lastdata != "")
                button_last_data_Click(sender, e); // load last selected data
            if (settings.lastmap != "")
                button_last_map_Click(sender, e); // load last selected map
            axGWControl1.InfoTooltipScrollable = true; // make ooltips scrollable
        }

        /// <summary>
        ///  
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void axGWControl1_MouseAction(object sender, _IGWControlEvents_MouseActionEvent e)
        {
            //if (e.action == (GWXLib.tagGWX_MouseAction)32)
            //{
            //  double lat = 0, lon = 0;
            //  axGWControl1.GetGeoCenter(out lat, out lon);
            //  int deb = 0;
            //}
            e.bHandled = 0;
        }

        private void LoadStrBtn_Click(object sender, EventArgs e)
        {
            AskFileName dlg = new AskFileName();
            string fn = dlg.GetFileName();
            if (fn != "")
                loadMap(fn);
        }

        private void button_load_coverage_Click(object sender, EventArgs e)
        {
            loadMap(Environment.CurrentDirectory + "\\..\\maps\\cover.cvr");
        }

        private void button_load_remote_coverage_Click(object sender, EventArgs e)
        {
            var gwsAddress = "http://109.124.101.250/";
            var cfgIniPath = Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "cfg.ini");
            foreach (var parts in File
                .ReadAllLines(cfgIniPath)
                .Where(line => line.StartsWith("GwsAddress="))
                .Select(line => line.Split(new[] {'='}))
                .Where(parts => parts.Length == 2))
            {
                gwsAddress = parts[1];
            }
            
            var dlg = new AskGwsKey();
            dlg.SetGwsUrlToDisplay(gwsAddress);
            var dlgResult = dlg.ShowDialog();
            if (dlgResult == DialogResult.Cancel)
                return;
            if (dlgResult == DialogResult.OK)
                gwsAddress += ";" + dlg.GwsKey;
            var fullLoadCommand = string.Format(" \t{0}\t{1}\\..\\maps\\cover.cvr", gwsAddress, Environment.CurrentDirectory);
            loadMap(fullLoadCommand);
        }

        private void button_unload_all_Click(object sender, EventArgs e)
        {
            var maps = axGWControl1.LoadedMaps;
            if (maps.rowCount() < 1)
                return;
            maps.moveFirst();
            do
            {
                var ma = (string) maps.getValue(0);
                axGWControl1.RemoveMap(ma);
            } while (maps.moveNext() > 0);
        }

        private void grid_jobs_MouseDoubleClick(object sender, MouseEventArgs e)
        {// double click on grid job - must show position of job point on map
            if (e.Button != MouseButtons.Left)
                return;
            int pointIx = JobIndexFromPoint(e.X, e.Y, true);
            if (pointIx == -1 || pointIx >= routeGeoPoints.Count)
                return;
            GeoPoint gp = routeGeoPoints[pointIx];

            axGWControl1.DeleteDBF("mark");
            axGWControl1.Table2Map("name='mark';descr='Marked Positions';structure='name text';", "m @502 'gwp.otl' 2", null);
            string coord = gp.Lon.ToString("g") + " " + gp.Lat.ToString("g");
            axGWControl1.ModifyTable("insert into [mark] set [name]='" + coord + "', metrics='" + "P E " + coord + ";" + "'", 0);

            axGWControl1.SetGeoCenter(gp.Lat, gp.Lon);
            axGWControl1.Refresh();
         }

    }
}

