﻿using System;
using System.Diagnostics;
using System.Windows;
using System.Windows.Controls;
using Newtonsoft.Json;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Data;
using System.Data.OleDb;
using System.Drawing;
using System.Linq;
using System.Net;
using System.Text;
using System.Windows.Forms;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Windows.Forms.Integration;
using System.Xml.Serialization;
using System.IO;
using DeliveryDemo;
using MapControl;
using Application = System.Windows.Forms.Application;
using Control = System.Windows.Forms.Control;
using DragDropEffects = System.Windows.Forms.DragDropEffects;
using DragEventArgs = System.Windows.Forms.DragEventArgs;
using ListView = System.Windows.Forms.ListView;
using ListViewItem = System.Windows.Forms.ListViewItem;
using MessageBox = System.Windows.Forms.MessageBox;
using Point = System.Drawing.Point;

namespace GWXDeliveryDemo
{
    public partial class FormGwsApi : Form
    {
        private ElementHost _ctrlHost;
        private Map _mapControl;
        private RoutesVisualManager _routeVisualManager;
        private DataSet input_data;
        private int cancontinue = 1;

        private string _sessionId;
        private DeliveryRequest _deliveryRequest;
        private DeliveryResult _deliveryResult;

        /// <summary>
        /// All program data that is stored at exit and restored at start up
        /// </summary>
        [Serializable]
        public struct Settings
        {
            public string lastmap;
            public string lastdata;
            public string optimize;
            public string distr;
            public bool multitrip;
            public bool fromgarage;
            public bool garagebetween;
            public bool togarage;
            public bool pointonce;
            public bool stacktype;
            public double intertrip;
            public bool groupdelivery;
            public bool usedate;
            public DateTime calcdate;
            public double addtime;
            public double addpath;
            public bool includetoroute;
            public bool multiday;
            public bool usetollroads;
        };

        private Settings settings;

        public struct GeoPoint
        {
            public double Lat;
            public double Lon;
        };

        public FormGwsApi()
        {
            InitializeComponent();

            settings.lastmap = "";
            settings.lastdata = "";
            settings.optimize = "optimizeTime";
            settings.distr = "favorableCombinations";
            settings.multitrip = false;
            settings.fromgarage = true;
            settings.togarage = true;
            settings.pointonce = true;
            settings.stacktype = true;
            settings.intertrip = 0;
            settings.groupdelivery = false;
            settings.addtime = 0;
            settings.addpath = 0;
            settings.includetoroute = false;
            settings.multiday = false;
            settings.usetollroads = true;
            SerializeSettings(false);

            _sessionId = GetSessionId();
        }

        private void SerializeSettings(bool serialize)
        {
            string filename = Application.ExecutablePath; // "*.exe"
            filename = filename.Substring(0, filename.Length - 3) + "settings"; // "*.exe" -> "*.settings"

            try
            {
                XmlSerializer settingsSerializer = new XmlSerializer(typeof(Settings));
                if (serialize)
                {
                    // saving
                    StreamWriter sw = new StreamWriter(filename);
                    settingsSerializer.Serialize(sw, settings);
                    sw.Close();
                }
                else
                {
                    // restoring
                    FileStream fs = new FileStream(filename, FileMode.Open);
                    settings = (Settings)settingsSerializer.Deserialize(fs);
                    fs.Close();
                }
            }
            catch (FileNotFoundException)
            {
                // The first start or file "DeliveryDemo.settings" deleted.
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            SerializeSettings(true);
        }

        /// <summary>
        /// slipts full file name in path and filename
        /// </summary>
        /// <param name="file">original filename</param>
        /// <param name="path">true to get path, false to get filename</param>
        /// <returns></returns>
        private static string DivideFilePath(string file, bool path)
        {
            for (int i = file.Length - 1; i >= 0; i--)
            {
                if (file[i] == '\\' || file[i] == '/' || file[i] == ':')
                    return path ? file.Substring(0, i) : file.Substring(i + 1);
            }
            return file;
        }

        /// <summary>
        /// returns path (path of DeliveryDemo.exe)\..\Data
        /// </summary>
        /// <returns></returns>
        private static string GetDefaultDataDirectory()
        {
            return DivideFilePath(DivideFilePath(Application.ExecutablePath, true), true) + @"\Data";
        }

        /// <summary>
        /// opens file dialog to get filename with specified filter
        /// </summary>
        /// <param name="filename">previous file name</param>
        /// <param name="filter">file filter i.e. "text files|*.txt"</param>
        /// <returns></returns>
        private static bool DoFileDialog(ref string filename, string filter)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = filter;
            if (filename != null && filename != "")
            {
                dlg.FileName = DivideFilePath(filename, false);
                dlg.InitialDirectory = DivideFilePath(filename, true);
            }
            else
                dlg.InitialDirectory = GetDefaultDataDirectory();

            if (dlg.ShowDialog() != DialogResult.OK)
                return false;
            filename = dlg.FileName;
            return true;
        }

        /// <summary>
        /// loads data from xml file (cars and jobs) 
        /// </summary>
        /// <param name="ask"></param>
        private void loadData(bool ask)
        {
            string filename = settings.lastdata;
            if (ask || filename == null || filename == "")
            {
                if (!DoFileDialog(ref filename, "xml and mdb files|*.xml;*.mdb"))
                    return;
            }
            WaitCursor wc = new WaitCursor();

            if (filename.ToLower().EndsWith(".mdb"))
            {
                try
                {
                    DataSet export_data = new DataSet(); // creating new

                    string ConnStr = @"Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" + filename + @";Persist Security Info=False";
                    OleDbConnection MyConn = new OleDbConnection(ConnStr);
                    MyConn.Open();
                    OleDbCommand Cmd = new OleDbCommand("select * from CARS", MyConn);
                    OleDbDataReader ObjReader = Cmd.ExecuteReader();

                    OleDbDataAdapter adapterCars = new OleDbDataAdapter("select * from CARS", MyConn);
                    adapterCars.Fill(export_data, "Cars");
                    OleDbDataAdapter adapterOrders = new OleDbDataAdapter("select * from ORDERS", MyConn);
                    adapterOrders.Fill(export_data, "Orders");
                    OleDbDataAdapter adapterZones = new OleDbDataAdapter("select * from ZONES", MyConn);
                    adapterZones.Fill(export_data, "Zones");

                    // where to save xml data?

                    SaveFileDialog dlg = new SaveFileDialog();
                    dlg.Filter = "xml files|*.xml";
                    dlg.InitialDirectory = GetDefaultDataDirectory();
                    dlg.DefaultExt = "xml";
                    dlg.Title = "Select a new xml file name to save the data";
                    //dlg.CheckPathExists = true;

                    if (dlg.ShowDialog() != DialogResult.OK)
                    {
                        input_data = export_data;
                        return;
                    }
                    string xmlfilename = dlg.FileName;

                    export_data.WriteXml(xmlfilename, XmlWriteMode.WriteSchema);

                    MyConn.Close();
                    filename = xmlfilename;
                }
                catch (Exception e)
                {
                    MessageBox.Show("Error exporting data from '" + filename + "':\r\n\r\n" + e.Message);
                    return;
                }
            }

            settings.lastdata = filename;

            input_data = new DataSet(); // creating new
            try
            {
                input_data.ReadXml(settings.lastdata, XmlReadMode.ReadSchema);
                dataGridViewCars.DataSource = input_data.Tables["Cars"];
                dataGridJobs.DataSource = input_data.Tables["Orders"];
                dataGridViewZones.DataSource = input_data.Tables["Zones"];
            }
            catch (Exception exc)
            {
                MessageBox.Show("File '" + settings.lastdata + "':\r\n\r\n" + exc.Message);
                input_data = null; // data object is invalid
            }
            ShowDataOnMap();
        }

        private void button_load_data_Click(object sender, EventArgs e)
        {
            loadData(true);
        }

        /// <summary>
        /// shows if parameters are set programmically, not by user
        /// </summary>
        private bool filling_settings = false;

        /// <summary>
        /// fills variants and sets controls from settings
        /// </summary>
        private void SetDeliveryPropertiesControls()
        {
            lb_variants.Items.Clear();

            var roadVariants = GetRoadVariants();

            if (roadVariants == null)
                return;

            lb_variants.Items.Add("— — — Available width — — —");
            foreach (var item in roadVariants.AvailableWidth)
            {
                lb_variants.Items.Add(item);
            }

            lb_variants.Items.Add("— — — Available height — — —");
            foreach (var item in roadVariants.AvailableHeight)
            {
                lb_variants.Items.Add(item);
            }

            lb_variants.Items.Add("— — — Available weight — — —");
            foreach (var item in roadVariants.AvailableWeight)
            {
                lb_variants.Items.Add(item);
            }

            lb_variants.Items.Add("— — — Available priority — — —");
            foreach (var item in roadVariants.AvailablePriority)
            {
                lb_variants.Items.Add(item);
            }

            lb_variants.Items.Add("— — — Available cargo — — —");
            foreach (var item in roadVariants.AvailableCargo)
            {
                lb_variants.Items.Add(item);
            }

            lb_variants.Items.Add("— — — Available permit — — —");
            foreach (var item in roadVariants.AvailablePermit)
            {
                lb_variants.Items.Add(item);
            }

            //GWX_VariantSubject[] subjs = new GWX_VariantSubject[]
            //{
            //    GWX_VariantSubject.GWX_VariantWidth,
            //    GWX_VariantSubject.GWX_VariantHeight, GWX_VariantSubject.GWX_VariantWeight,
            //    GWX_VariantSubject.GWX_VariantPriority,
            //    GWX_VariantSubject.GWX_VariantCargo, GWX_VariantSubject.GWX_VariantLicence
            //};

            //foreach (GWX_VariantSubject subj in subjs)
            //{
            //    lb_variants.Items.Add("— — — " + subj.ToString() + " — — —");
            //    GWStringList list = delivery.GetVariants(subj);
            //    for (int i = 0; i < list.ItemCount; i++)
            //        lb_variants.Items.Add(list.GetItem(i));
            //}

            // settings -> controls
            filling_settings = true;
            cb_optimization.Text = settings.optimize;
            cb_distribution.Text = settings.distr;
            xb_multitrip.Checked = settings.multitrip;
            xb_fromgarage.Checked = settings.fromgarage;
            xb_garagebetween.Checked = settings.garagebetween;
            xb_togarage.Checked = settings.togarage;
            xb_pointonce.Checked = settings.pointonce;
            xb_groupdelivery.Checked = settings.groupdelivery;
            xb_stack.Checked = settings.stacktype;
            tb_interval.Text = settings.intertrip.ToString();
            tb_interval.Modified = false;
            xb_date.Checked = settings.usedate;
            dtp_date.Enabled = settings.usedate;
            dtp_date.Value = settings.usedate ? settings.calcdate : DateTime.Today;
            tb_addtime.Text = settings.addtime.ToString();
            tb_addtime.Modified = false;
            tb_addpath.Text = settings.addpath.ToString();
            tb_addpath.Modified = false;
            xb_includeadditionals.Checked = settings.includetoroute;
            xb_multiday.Checked = settings.multiday;
            xb_usetollroads.Checked = settings.usetollroads;
            filling_settings = false;
        }

        /// <summary>
        /// user changed some delivery parameter
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void on_delivery_setting_change(object sender, EventArgs e)
        {
            if (filling_settings)
                return;
            if (sender == cb_optimization) settings.optimize = cb_optimization.Text;
            else if (sender == cb_distribution) settings.distr = cb_distribution.Text;
            else if (sender == xb_multitrip) settings.multitrip = xb_multitrip.Checked;
            else if (sender == xb_fromgarage) settings.fromgarage = xb_fromgarage.Checked;
            else if (sender == xb_garagebetween) settings.garagebetween = xb_garagebetween.Checked;
            else if (sender == xb_togarage) settings.togarage = xb_togarage.Checked;
            else if (sender == xb_pointonce) settings.pointonce = xb_pointonce.Checked;
            else if (sender == xb_stack) settings.stacktype = xb_stack.Checked;
            else if (sender == xb_groupdelivery) settings.groupdelivery = xb_groupdelivery.Checked;
            else if (sender == xb_includeadditionals) settings.includetoroute = xb_includeadditionals.Checked;
            else if (sender == xb_multiday) settings.multiday = xb_multiday.Checked;
            else if (sender == xb_usetollroads) settings.usetollroads = xb_usetollroads.Checked;
            else if (sender == xb_date)
            {
                settings.usedate = xb_date.Checked;
                dtp_date.Enabled = settings.usedate;
            }
            else if (sender == dtp_date) settings.calcdate = dtp_date.Value;
        }

        /// <summary>
        /// when tb_interval loses focus and is modified by user, applies new value
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void tb_interval_Leave(object sender, EventArgs e)
        {
            if (!tb_interval.Modified)
                return;
            try
            {
                settings.intertrip = Convert.ToDouble(tb_interval.Text);
            }
            catch (Exception)
            {
            }
            tb_interval.Text = settings.intertrip.ToString();
            tb_interval.Modified = false;
        }

        private void tb_addtime_Leave(object sender, EventArgs e)
        {
            if (!tb_addtime.Modified)
                return;
            try
            {
                settings.addtime = Convert.ToDouble(tb_addtime.Text);
            }
            catch (Exception)
            {
            }
            tb_addtime.Text = settings.addtime.ToString();
            tb_addtime.Modified = false;
        }

        private void tb_addpath_Leave(object sender, EventArgs e)
        {
            if (!tb_addpath.Modified)
                return;
            try
            {
                settings.addpath = Convert.ToDouble(tb_addpath.Text);
            }
            catch (Exception)
            {
            }
            tb_addpath.Text = settings.addpath.ToString();
            tb_addpath.Modified = false;
        }

        private void button_last_data_Click(object sender, EventArgs e)
        {
            loadData(false);
        }

        /// <summary>
        /// shows on map cars and jobs from input_data
        /// </summary>
        private void ShowDataOnMap()
        {
            
            if (input_data == null)
                return;
            /*
            if (axGWControl1.MapName == "")
                return;
            axGWControl1.DeleteDBF("cars");
            axGWControl1.DeleteDBF("jobs");
            axGWControl1.DeleteDBF("joblines");
            axGWControl1.DeleteDBF("zones");

            axGWControl1.Table2Map("name='cars';descr='Cars';structure='name text Car name';", "m @502 'gwp.otl' 69",
                                   null);
            axGWControl1.Table2Map("name='jobs';descr='Jobs';structure='name text Job name, img integer image';",
                                   "m @500 'gwp.otl' [img]+65", null);
            axGWControl1.Table2Map("name='joblines';descr='Jobs';structure='name text Job name';", "l @501 red 1", null);

            axGWControl1.Table2Map("name='zones';descr='Zones';structure='id text Zone id, descr text Description, color integer Zone color';",
                "a @501 [color] 35; p @502 [color] 2", //"l @501 red 3",
                null);
            //*/
            RemoveCarLines();
            ShowCars(new MyReader(new DataTableReader(input_data.Tables["Cars"])));
            ShowJobs(new MyReader(new DataTableReader(input_data.Tables["Orders"])));
            //ShowZones(new MyReader(new DataTableReader(input_data.Tables["Zones"])));
            //axGWControl1.Refresh();
        }

        /// <summary>
        /// Shows cars on map
        /// </summary>
        /// <param name="mCars"></param>
        private void ShowCars(MyReader mCars)
        {
            while (mCars.Read())
            {
                double lon = mCars.GetDouble("XCOORD"), lat = mCars.GetDouble("YCOORD");
                string m = "P E " + lon.ToString("g") + " " + lat.ToString("g") + ";";
                // metrics requires (X Y), i.e. (Lon Lat)
                //axGWControl1.ModifyTable(
                  //  "insert into [cars] set [name]='" + mCars.GetString("N") + "', metrics='" + m + "'", 0);
            }
        }

        /// <summary>
        /// Shows jobs on map
        /// </summary>
        /// <param name="mJobs"></param>
        private void ShowJobs(MyReader mJobs)
        {
            while (mJobs.Read())
            {
                string jobName = mJobs.GetString("N");
                double lat0 = mJobs.GetDouble("GETYCOORD"), lon0 = mJobs.GetDouble("GETXCOORD");
                double lat1 = mJobs.GetDouble("PUTYCOORD"), lon1 = mJobs.GetDouble("PUTXCOORD");
                string pos0 = lon0.ToString("g") + " " + lat0.ToString("g");
                string pos1 = lon1.ToString("g") + " " + lat1.ToString("g");

                var id = (int)jobName.GetHashCode();
                var locations = new [] {new Location(lat0, lon0), new Location(lat1, lon1)};
                _routeVisualManager.AddSegment(id, locations, "Заказ №: " + jobName);
                _routeVisualManager.AddStop(id, new Location(lat0, lon0), "", "");
                _routeVisualManager.AddStop(id, new Location(lat1, lon1), "", "");


                //axGWControl1.ModifyTable(
                //    "insert into [jobs] set [name]='" + jobName + "', [img]=0, metrics='P E " + pos0 + ";'", 0);
                //axGWControl1.ModifyTable(
                //    "insert into [jobs] set [name]='" + jobName + "', [img]=1, metrics='P E " + pos1 + ";'", 0);
                //axGWControl1.ModifyTable(
                //    "insert into [joblines] set [name]='" + jobName + "', metrics='L E " + pos0 + ", " + pos1 + ";'", 0);
            }
        }

        /// <summary>
        /// Shows zones on map
        /// </summary>
        /// <param name="mJobs"></param>
        private void ShowZones(MyReader mZones)
        {// "name='zones';descr='Zones';structure='ID text Zone id, descr text Description';"
            while (mZones.Read())
            {
                string id = mZones.GetString("ID");
                string descr = mZones.GetString("DESCR");
                string metrics = mZones.GetString("METRICS");
                int color = mZones.GetInt32("COLOR");
                //if (metrics != "")
                //    axGWControl1.ModifyTable(
                //        "insert into [zones] set [id]='" + id + "', [descr]='" + descr + "', color=" + color.ToString() +
                //        ", metrics='" + metrics + "'", 0);
            }
        }

        private void button_calc_Click(object sender, EventArgs e)
        {
            Calculate();

            button_calc.Enabled = true;
            button_abort.Enabled = false;
            progress_calc.Value = 0;
        }

        /// <summary>
        /// loads cars and jobs into delivery, makes delivery to calculate routes, gets results
        /// </summary>
        private void Calculate()
        {
            WaitCursor wc;
            try
            {
                if (input_data == null)
                    throw new Exception("Data not loaded");
            }
            catch (Exception exc)
            {
                MessageBox.Show(exc.Message);
                return;
            }

            _deliveryRequest = new DeliveryRequest();
            _deliveryRequest.SessionId = uint.Parse(_sessionId);
            _deliveryRequest.AccessKey = AccessKeyTextBox.Text;

            button_abort.Enabled = true;
            button_calc.Enabled = false;
            cancontinue = 1;

            // send data to delivery
            SetDeliveryProperties();
            AddCarsToDelivery();
            AddJobsToDelivery();
            AddZonesToDelivery();

            var res = GetDelivery(_deliveryRequest);
            if (string.IsNullOrEmpty(res))
            {
                MessageBox.Show("Empty result. Please, check data or access key.");
                return;
            }
            try
            {
                var errorResult = JsonConvert.DeserializeObject<List<string>>(res);
                MessageBox.Show(string.Format("Error: {0}", res));
                return;
            }
            catch (Exception)
            {
            }
            _deliveryResult = JsonConvert.DeserializeObject<DeliveryResult>(res);

            // filling error list
            //GWStringList errors = delivery.GetErrors();
            //string err_text = "";
            //for (int pos = errors.moveFirst(); pos != 0; pos = errors.moveNext())
            //    err_text += errors.Item + "\r\n";
            //tb_errors.Text = err_text;

            // show total result on status bar
            var tot_cnt = _deliveryResult.Orders.Count(o => o.CarId != 0);
            var not_cnt = _deliveryResult.Orders.Count(o => o.CarId == 0);
            var tot_len = _deliveryResult.Cars.Sum(c => c.TotalRouteLength) / 1000.0;
            var tot_time = _deliveryResult.Cars.Sum(c => c.TotalRouteTime);

            var res_txt = "Length=" + tot_len.ToString("0.#") + " km, Time=" + MinutesToStrring(tot_time) +
                             "', Jobs: " + tot_cnt.ToString() + ", ";
            if (not_cnt == 0) res_txt += "all resolved";
            else res_txt += not_cnt.ToString() + " not resolved";
            toolStripStatusLabel2.Text = res_txt;
            FillCarsList(true);
            if (list_cars.Items.Count > 0)
                list_cars.Items[0].Selected = true;
        }

        /// <summary>
        /// set current settings to delivery
        /// </summary>
        private void SetDeliveryProperties()
        {
            var parameters = _deliveryRequest.DeliveryData.Parameters;

            parameters.OptimizationMode = settings.optimize;
            parameters.DistributionMethod = settings.distr;
            parameters.AllowMultiTripRoutes = settings.multitrip;
            parameters.CountPathFromGarage = settings.fromgarage;
            parameters.AllowGarageInMultiTrip = settings.garagebetween;
            parameters.CountPathToGarage = settings.togarage;
            parameters.VisitEachPointOnlyOnce = settings.pointonce;
            parameters.UseStackUnloading = settings.stacktype;
            parameters.MultiTripTimeInterval = (uint)(settings.intertrip * 60);
            parameters.IsGroupDelivery = settings.groupdelivery;
            parameters.AdditionalTimeInPoint = (int)(settings.addtime * 60);
            parameters.AdditionalDistanceInPoint = (int)(settings.addpath * 1000);
            parameters.IncludeAdditionsInRoute = settings.includetoroute;
            parameters.MultiDayRoutes = settings.multiday;
            parameters.UseTollRoads = settings.usetollroads;
        }

        //private void axGWControl1_AsyncReqComplete(object sender, AxGWXLib._IGWControlEvents_AsyncReqCompleteEvent e)
        //{
        //    String s = e.response;
        //}

        private void button_abort_Click(object sender, EventArgs e)
        {
            cancontinue = 0;
        }

        private DateTime ZeroDate = DateTime.MinValue;
        private DateTime EmptyDate = new DateTime(1899, 12, 30);
        private DateTime CheckDefaultDate(DateTime dt, DateTime dt_prev)
        { // dt = {30.12.1899 1:05:00}
            if (!settings.usedate)
                return dt;
            if (dt.Date != EmptyDate.Date)
                return dt;
            dt = settings.calcdate.Date + dt.TimeOfDay;
            if (dt_prev > dt)
                dt.AddDays(1);
            return dt;
        }

        /// <summary>
        /// fills car table in delivery from input_data
        /// </summary>
        private void AddCarsToDelivery()
        {
            MyReader mCars = new MyReader(new DataTableReader(input_data.Tables["Cars"]));
            while (mCars.Read())
            {
                var car = new DeliveryCar();

                car.Id = (uint)mCars.GetString("N").GetHashCode();

                car.MaxWeight = (uint)mCars.GetDouble("MAXWEIGHT");
                car.MaxVolume = (uint)mCars.GetDouble("MAXVOLUME");
                car.MaxVolume2 = (uint)mCars.GetDouble("MAXVOLUME2");
                car.MaxCost = (uint)mCars.GetDouble("MAXCOST");
                car.MaxOrdersCount = (uint)mCars.GetInt32("MAXORD");
                car.MaxRouteLength = (uint)(mCars.GetDouble("MAXLEN") * 60);
                car.MaxRouteTime = (uint)(mCars.GetDouble("MAXTIME") * 60.0); // hours -> mins

                car.SelfWeight = (uint)mCars.GetDouble("SELFWEIGHT");
                car.InitialPosition = new GWXDeliveryDemo.GeoPoint();
                car.InitialPosition.Latitude = mCars.GetDouble("YCOORD");
                car.InitialPosition.Longtitude = mCars.GetDouble("XCOORD");

                DateTime dt1, dt2;
                if (mCars.GetDateTime("WRKBEG", out dt1)) car.WorkStartTime = CheckDefaultDate(dt1, ZeroDate).ToString("yyyy-MM-ddTHH:mm");
                if (mCars.GetDateTime("WRKEND", out dt2)) car.WorkFinishTime = CheckDefaultDate(dt2, dt1).ToString("yyyy-MM-ddTHH:mm");

                if (mCars.GetDateTime("LUNCHBEG", out dt1)) car.LunchStartTime = dt1.ToString("HH:mm"); // date not used
                if (mCars.GetDateTime("LUNCHEND", out dt1)) car.LunchFinishTime = dt1.ToString("HH:mm");

                car.Width = (uint)mCars.GetDouble("WIDTH");
                car.Height = (uint)mCars.GetDouble("HEIGHT");
                car.CargoType = (uint)(mCars.GetDouble("CARGO")*10);
                car.Permit = mCars.GetString("LICENCE");

                car.CarType = (uint)mCars.GetString("CARTYPE").GetHashCode();
                car.FuelConsumptionPerKm = mCars.GetDouble("FUELPERKM");

                car.FillZoneListFromString(mCars.GetString("LOADZONES"), true);
                car.FillZoneListFromString(mCars.GetString("UNLOADZONES"), false);

                car.WorkTime = mCars.GetString("WORKTIME");
                car.UseTollRoads = mCars.GetInt32("USETOLLROARDS", -1);

                _deliveryRequest.DeliveryData.Cars.Add(car);
            }
        }

        /// <summary>
        /// fills job table in delivery from input_data
        /// </summary>
        private void AddJobsToDelivery()
        {
            MyReader mJobs = new MyReader(new DataTableReader(input_data.Tables["Orders"]));
            while (mJobs.Read())
            {

                var job = new DeliveryOrder();

                job.Id = (uint)mJobs.GetString("N").GetHashCode();
                job.Priority =(uint)mJobs.GetInt32("PRIORITY");

                job.Weight = (uint)mJobs.GetDouble("WEIGHT");
                job.Volume = (uint)mJobs.GetDouble("VOLUME");
                job.CarVolumeUsage = (uint)mJobs.GetInt32("VOLUME_DISP");
                job.Cost = (uint)mJobs.GetDouble("COST");
                job.FillCarTypesFromString(mJobs.GetString("CARTYPES"));
                job.MaxTransportationTime = (uint)(mJobs.GetDouble("MAXTRANSPTIME") * 60.0);//// already in minutes *60.0; // hours -> mins
                job.LoadPosition = new GWXDeliveryDemo.GeoPoint
                {
                    Latitude = mJobs.GetDouble("GETYCOORD"),
                    Longtitude = mJobs.GetDouble("GETXCOORD")
                };
                
                DateTime dt1, dt2;
                if (mJobs.GetDateTime("GETBEG", out dt1)) job.LoadStartTime = CheckDefaultDate(dt1, ZeroDate).ToString("yyyy-MM-ddTHH:mm");
                if (mJobs.GetDateTime("GETEND", out dt2)) job.LoadFinishTime = CheckDefaultDate(dt2, dt1).ToString("yyyy-MM-ddTHH:mm");
                job.LoadDuration = (uint)(mJobs.GetDouble("GETDUR") * 60);

                job.HandOverPosition = new GWXDeliveryDemo.GeoPoint
                {
                    Latitude = mJobs.GetDouble("PUTYCOORD"),
                    Longtitude = mJobs.GetDouble("PUTXCOORD")
                };
                if (mJobs.GetDateTime("PUTBEG", out dt1)) job.HandOverStartTime = CheckDefaultDate(dt1, ZeroDate).ToString("yyyy-MM-ddTHH:mm");
                if (mJobs.GetDateTime("PUTEND", out dt2)) job.HandOverFinishTime = CheckDefaultDate(dt2, dt1).ToString("yyyy-MM-ddTHH:mm");
                job.HandOverDuration = (uint)(mJobs.GetDouble("PUTDUR")*60);

                if (mJobs.GetString("CARN").Length > 0)
                {
                    job.AttachedCarId = (uint) mJobs.GetString("CARN").GetHashCode();
                    job.LoadPointIndexInRoute = mJobs.GetInt32("RTGETIX");
                    job.HandOverPointIndexInRoute = mJobs.GetInt32("RTPUTIX");
                }

                job.LoadTime = mJobs.GetString("LOADTIME");
                job.HandOverTime = mJobs.GetString("UNLOADTIME");

                _deliveryRequest.DeliveryData.Orders.Add(job);
            }
        }

        /// <summary>
        /// fills zones table in delivery from input_data
        /// </summary>
        private void AddZonesToDelivery()
        {
            MyReader mZones = new MyReader(new DataTableReader(input_data.Tables["Zones"]));
            while (mZones.Read())
            {
                var zone = new DeliveryZone();
                zone.Id = (uint)mZones.GetString("ID").GetHashCode();
                zone.Metrics = mZones.GetString("METRICS");
                _deliveryRequest.DeliveryData.Zones.Add(zone);
            }
        }

        /// <summary>
        /// converts to string in format 26 -> "26", 156.5 -> "2:37"
        /// </summary>
        /// <param name="minutes"></param>
        /// <returns></returns>
        private static string MinutesToStrring(double minutes)
        {
            int m = (int)(minutes + 0.5);
            if (m < 60) return m.ToString();
            int h = m / 60;
            m = m % 60;
            return h.ToString() + ":" + m.ToString("00");
        }

        /// <summary>
        /// adds or updates item in car list
        /// </summary>
        /// <param name="firsttime">add new item</param>
        /// <param name="i">row index</param>
        /// <param name="name">car name</param>
        /// <param name="jobsCount">jobs count</param>
        /// <param name="length">route length</param>
        /// <param name="time">route time</param>
        /// <param name="errCount">count of errors</param>
        private ListViewItem AddToCarList(bool firsttime, ref int i, string name, int jobsCount, double length, double time,
                                  int errCount)
        {
            ListViewItem lwi = null;
            if (firsttime)
            {
                lwi = new ListViewItem(name);
                list_cars.Items.Add(lwi);
                for (int k = 0; k < 4; k++)
                    lwi.SubItems.Add("");
            }
            else
            {
                lwi = list_cars.Items[i++];
            }
            // name is the same
            lwi.SubItems[1].Text = jobsCount.ToString();
            lwi.SubItems[2].Text = length > 0.0 ? length.ToString("0.#") : "";
            lwi.SubItems[3].Text = time > 0.0 ? MinutesToStrring(time) : "";
            lwi.SubItems[4].Text = errCount > 0 ? errCount.ToString() : "";
            return lwi;
        }

        private string GetCarNameByHashId(uint id)
        {
            var mCars = new MyReader(new DataTableReader(input_data.Tables["Cars"]));
            while (mCars.Read())
            {
                var carName = mCars.GetString("N");
                if (id == (uint) carName.GetHashCode())
                    return carName;
            }
            return "";
        }

        private string GetOrderNameByHashId(uint id)
        {
            var mJobs = new MyReader(new DataTableReader(input_data.Tables["Orders"]));
            while (mJobs.Read())
            {
                var jobName = mJobs.GetString("N");
                if (id == (uint) jobName.GetHashCode())
                    return jobName;
            }
            return "";
        }

        private string GetOrderDescriptionByHashId(uint id, bool isLoadingPoint)
        {
            var mJobs = new MyReader(new DataTableReader(input_data.Tables["Orders"]));
            while (mJobs.Read())
            {
                var jobName = mJobs.GetString("N");
                if (id != (uint)jobName.GetHashCode())
                    continue;
                return isLoadingPoint ? mJobs.GetString("GETDESCR") : mJobs.GetString("PUTDESCR");
            }
            return "";
        }

        private void GetOrderParamsByHashId(uint id, out double weight, out double volume, out double cost)
        {
            weight = volume = cost = 0;
            var mJobs = new MyReader(new DataTableReader(input_data.Tables["Orders"]));
            while (mJobs.Read())
            {
                var jobName = mJobs.GetString("N");
                if (id != (uint) jobName.GetHashCode()) 
                    continue;
                weight = mJobs.GetDouble("WEIGHT");
                volume = mJobs.GetDouble("VOLUME");
                //job.CarVolumeUsage = (uint)mJobs.GetInt32("VOLUME_DISP");
                cost = mJobs.GetDouble("COST");
                return;
            }
        }

        private string GetZoneNameByHashId(uint id)
        {
            var mZones = new MyReader(new DataTableReader(input_data.Tables["Zones"]));
            while (mZones.Read())
            {
                var zoneName = mZones.GetString("ID");
                if (id == (uint) zoneName.GetHashCode())
                    return zoneName;
            }
            return "";
        }

        /// <summary>
        /// Fills result car list + items "all" and "not resolved"
        /// </summary>
        /// <param name="firsttime">true to add items for the first time, false for update existing items</param>
        private void FillCarsList(bool firsttime) // firsttime - fill list, else update items
        {
            if (firsttime)
                list_cars.Items.Clear();
            var tot_cnt = _deliveryResult.Orders.Count(o => o.CarId != 0);
            var not_cnt = _deliveryResult.Orders.Count(o => o.CarId == 0);
            var tot_len = _deliveryResult.Cars.Sum(c => c.TotalRouteLength) / 1000.0;
            var tot_time = _deliveryResult.Cars.Sum(c => c.TotalRouteTime);

            int ii = 0;
            //// summary
            AddToCarList(firsttime, ref ii, "< — all — >", tot_cnt, tot_len, tot_time, 0);
            int errCnt = 0;
            foreach (var car in _deliveryResult.Cars)
            {
                AddToCarList(firsttime, ref ii, GetCarNameByHashId(car.CarId), (int)car.TotalOrdersCount, car.TotalRouteLength / 1000.0, (int)car.TotalRouteTime,
                             (int)car.ErrorPointsCount);
                errCnt += (int)car.ErrorPointsCount;
            }
            // not distributed
            ListViewItem lwi = AddToCarList(firsttime, ref ii, "<not resolved>", not_cnt, 0.0, 0.0, 0);
            if (errCnt > 0)
                lwi.SubItems[4].Text = errCnt.ToString();
        }

        private const string ROUTETABLEHEADER = "Route";
        private const string POINTSHEADER = "Points of car ";

        /// <summary>
        /// Removes all lines of cars from map
        /// </summary>
        private void RemoveCarLines()
        {
            _routeVisualManager.Clear();
        }

        /// <summary>
        /// shows lines of route for specified car
        /// </summary>
        /// <param name="car"></param>
        private void ShowCarLines(DeliveryCarResult car, bool point_times)
        {
            _routeVisualManager.Clear();
            string name = GetCarNameByHashId(car.CarId);
            
            var previousMainPointName = "";
            var segmentPoints = new List<Location>();

            foreach (var routePoint in car.Route)
            {
                var location = new Location(routePoint.Coordinates.Latitude, routePoint.Coordinates.Longtitude);
                if (routePoint.Type == 10)
                {
                    segmentPoints.Add(location);
                    continue;
                }

                var currentDescription = routePoint.Description;
                if (string.IsNullOrEmpty(currentDescription) && routePoint.OrderId.HasValue)
                {
                    currentDescription = GetOrderDescriptionByHashId(routePoint.OrderId.Value, routePoint.Type == 1);
                }
                var segmentText = string.Format("{0} -> {1}", previousMainPointName, currentDescription);
                var pointText = GetRoutePointTooltipText(routePoint);
                if (segmentPoints.Count > 0)
                {
                    _routeVisualManager.AddSegment((int) car.CarId, segmentPoints, segmentText);
                    segmentPoints.Clear();
                }
                _routeVisualManager.AddStop((int)car.CarId, location, pointText, pointText);
                previousMainPointName = currentDescription;
            }
            
        }

        string GetRoutePointTooltipText(DeliveryResultPoint routePoint)
        {
            var orderName = "";
            if (routePoint.OrderId != null)
            {
                orderName = GetOrderNameByHashId(routePoint.OrderId.Value);
            }
            switch (routePoint.Type)
            {
                case 0:
                    return string.Format("Старт маршрута\nТочка: {0}\nВремя: {1}", routePoint.Description, routePoint.ArrivalTime);
                case 1:
                    return string.Format("Точка: {0}\nЗагрузка: заказ №{1},\nВремя: {2}", routePoint.Description,
                        orderName, routePoint.ArrivalTime);
                case 2:
                    return string.Format("Точка: {0}\nВыгрузка: заказ №{1},\nВремя: {2}", routePoint.Description,
                        orderName, routePoint.ArrivalTime);
                case 3:
                    return string.Format("Конец маршрута\nТочка: {0}\nВремя: {1}", routePoint.Description, routePoint.ArrivalTime);
                case 4:
                    return string.Format("Конец рейса\nТочка: {0}\nВремя: {1}", routePoint.Description, routePoint.ArrivalTime);
                default:
                    return "";
            }
        }

        /// <summary>
        /// adds column headers with specified text and width
        /// </summary>
        /// <param name="colheaders">like "Name:90;Car:90;Weight:50;"</param>
        private void PrepareJobGridColumns(string colheaders) // "Name:90;Car:90;Weight:50;"
        {
            DataGridViewCell cell = new DataGridViewTextBoxCell();
            foreach (string coldef in colheaders.Split(';'))
            {
                int sep = coldef.IndexOf(':');
                if (sep < 0) continue;
                DataGridViewColumn col = new DataGridViewColumn();
                col.HeaderText = coldef.Substring(0, sep);
                col.Width = Convert.ToInt32(coldef.Substring(sep + 1));
                col.CellTemplate = cell;
                grid_jobs.Columns.Add(col);
            }
        }

        /// <summary>
        /// means of point types
        /// </summary>
        private static string[] point_types = { "begin", "load", "unload", "end", "trip end", "not distributed", "lunch", "interval", "", "", ""};

        /// <summary>
        /// shows current value and its increment with + or - when delta !=0
        /// </summary>
        /// <param name="value">current value</param>
        /// <param name="delta">increment</param>
        /// <returns></returns>
        private static string ValueAndDelta(double value, double delta)
        {
            string fmt = delta > 0.0 ? "{0:0.#} (+{1:0.#})" : delta < 0.0 ? "{0:0.#} ({1:0.#})" : "{0:0.#}";
            return string.Format(fmt, value, delta);
        }

        /// <summary>
        /// means of error numbers
        /// </summary>
        private static string[] error_means =
        {
            "Error", "Point visited more than once", "Max weight is exceeded by {0}", "Max volume is exceeded by {0}",
            "Max cost is exceeded by {0}", "Max order number is exceeded by {0}", "Car type mismatch",
            "Handling time exceeded by {0}",
            "Earlier by {0}", "Later by {0}", "Max duration exceeded by {0}", "Max length exceeded by {0}",
            "Work ends later by {0}",
            "Zone condition is violated",
            "Max volume 2 is exceeded by {0}", //#14#
            "Different groups" //"#15#"
        };

        /// <summary>
        /// converts error string returned in route table
        /// </summary>
        /// <param name="errmsg">like "#nr#value#nr2#value2"</param>
        /// <returns></returns>
        private static string TranslateError(string errmsg)
        {
            // #nr#value#nr2#value2 
            string res = "";
            int pos0;
            while ((pos0 = errmsg.IndexOf('#')) >= 0)
            {
                pos0++;
                int pos1 = errmsg.IndexOf('#', pos0);
                if (pos1 < pos0)
                    break;
                int nr = 0;
                try
                {
                    nr = Convert.ToInt32(errmsg.Substring(pos0, pos1 - pos0));
                }
                catch (Exception)
                {
                    break;
                }

                pos0 = pos1 + 1;
                pos1 = errmsg.IndexOf('#', pos0);
                if (pos1 < 0) pos1 = errmsg.Length;

                if (res != "") res += " / ";
                res += string.Format(error_means[nr], errmsg.Substring(pos0, pos1 - pos0));
                errmsg = errmsg.Substring(pos1);
            }
            return res;
        }

        /// <summary>
        /// in different modes to fill job list, dependiong on selected item in car list, where is column with job name
        /// </summary>

        private void list_cars_SelectedIndexChanged(object sender, EventArgs e)
        {
            FillJobList();
        }


        /// <summary>
        /// fills job list in different modes depending on what is selected in car list - car, all or 'not distributed jobs' item
        /// </summary>
        private void FillJobList()
        {
            carOfJobList = "";
            routeGeoPoints.Clear();
            ListView.SelectedIndexCollection ixs = list_cars.SelectedIndices;
            int ix = -1;
            if (ixs.Count == 1) ix = ixs[0];
            RemoveCarLines();
            grid_jobs.Rows.Clear();
            grid_jobs.Columns.Clear();
            lb_route_descr.Items.Clear();
            if (ix < 0)
                return;
            if (ix == 0)
            {
                // all cars
                foreach (var car in _deliveryResult.Cars)
                    ShowCarLines(car, false);

                PrepareJobGridColumns("Job:60;Car:60;Weight:50;Volume:50;Cost:45;");
                foreach (var job in _deliveryResult.Orders)
                {
                    var carName = GetCarNameByHashId(job.CarId);
                    if (carName != "")
                    {
                        double weight, volume, cost;
                        GetOrderParamsByHashId(job.OrderId, out weight, out volume, out cost);
                        grid_jobs.Rows.Add(GetOrderNameByHashId(job.OrderId), carName, weight, volume, cost);
                    }
                }
                jobNameColumn = 0;
            }
            else if (ix == list_cars.Items.Count - 1) // list_cars.Items.Count-1 - <not resolved orders>
            {
                PrepareJobGridColumns("Job:60;Weight:50;Volume:50;Cost:45;");
                foreach (var job in _deliveryResult.Orders)
                {
                    if (GetCarNameByHashId(job.CarId) == "")
                    {
                        double weight, volume, cost;
                        GetOrderParamsByHashId(job.OrderId, out weight, out volume, out cost);
                        grid_jobs.Rows.Add(GetOrderNameByHashId(job.OrderId), weight, volume, cost);
                    }
                }
                jobNameColumn = 0;
            }
            else
            {
                // only one car
                var carName = list_cars.Items[ix].Text;
                var carResult = _deliveryResult.Cars.Single(car => car.CarId == (uint) carName.GetHashCode());
                ShowCarLines(carResult, true);
                FillJobGrid(carResult);
                FillRouteList(carResult);
            }
        }

        private void FillJobGrid(DeliveryCarResult carResult)
        {
            ///* Table columns:
            //    0  "Type",     "INTEGER", "Point type"
            //    1  "Lat",      "DOUBLE",  "Latitude"
            //    2  "Lon",      "DOUBLE",  "Longitude"
            //    3  "Time",     "DOUBLE",  "Arrival time"
            //    4  "TimeText", "TEXT",    "Arrival time as text"
            //    5  "Weight",   "DOUBLE",  "Job weight"
            //    6  "Volume",   "DOUBLE",  "Job volume"
            //    7  "Cost",     "DOUBLE",  "Job cost"
            //    8  "Job",      "TEXT",    "Job name"
            //    9  "Error",    "TEXT",    "Error message"
            //    10 "Length",   "DOUBLE",  "Path length"
            //    11 "Duration", "DOUBLE",  "Path duration"
            //    12 "Path",     "TABLE",   "Path as table"
            //    13 "Metrics",  "TEXT",    "Path as metrics string"
            //    14 "Description", "TEXT", "Point description"
            //    15 "VolumeDisp",  "INTEGER", "Volume 1 or 2"
            // */
            PrepareJobGridColumns(
                "Type:40;Job:60;Description:80;Time:50;Weight:70;Volume:100;Cost:50;Error:120");
                //"Type:40;Job:60;Description:80;Time:50;Weight:70;Volume:100;Cost:50;Length:50;Time:50;Error:120");
            //                     0       8,     14             4       5         6         7       10        11      9
            double sum_weight = 0, sum_volume1 = 0, sum_volume2 = 0, sum_cost = 0;
            foreach (var resultPoint in carResult.Route)
            {
                if (resultPoint.Type == 10)
                    continue;
                double weight = 0, volume1 = 0, cost = 0, volume2 = 0;
                var volume_disp = resultPoint.VolumeUsed.HasValue? resultPoint.VolumeUsed.Value : 0;
                
                GetOrderParamsByHashId(resultPoint.OrderId.Value, out weight, out volume1, out cost);
                var ptype = resultPoint.Type;
                var orderName = GetOrderNameByHashId(resultPoint.OrderId.Value);
                
                if (volume_disp == 2)
                {
                    volume2 = volume1;
                    volume1 = 0;
                }
                
                if (ptype == 2)
                {
                    weight = -weight;
                    cost = -cost;
                    volume1 = -volume1;
                    volume2 = -volume2;
                }
                sum_weight += weight;
                sum_volume1 += volume1;
                sum_volume2 += volume2;
                sum_cost += cost;

                var gridIx = grid_jobs.Rows.Add(
                    // type, job name, point_descr, time str, 
                    point_types[ptype], 
                    orderName, 
                    GetOrderDescriptionByHashId(resultPoint.OrderId.Value, ptype == 1),
                    CheckDefaultDate(resultPoint.ArrivalTime, ZeroDate),
                    // weight, volume, cost
                    ValueAndDelta(sum_weight, weight),
                    ValueAndDelta(sum_volume1, volume1) + " / " + ValueAndDelta(sum_volume2, volume2),
                    ValueAndDelta(sum_cost, cost),
                    // length, time
                    //((double)resultPoint.).ToString("0.#"),
                    //MinutesToStrring((double)tbl.getValue(11)), 
                    // error
                    TranslateError(resultPoint.ErrorMessage)
                );

                GeoPoint gp;
                gp.Lat = resultPoint.Coordinates.Latitude;
                gp.Lon = resultPoint.Coordinates.Longtitude;
                routeGeoPoints.Add(gp);
            }
            
            jobNameColumn = 1;
        }

        public static String StringEncodingConvert(String strText, String strSrcEncoding, String strDestEncoding)
        {
            System.Text.Encoding srcEnc = System.Text.Encoding.GetEncoding(strSrcEncoding);
            System.Text.Encoding destEnc = System.Text.Encoding.GetEncoding(strDestEncoding);
            byte[] bData = srcEnc.GetBytes(strText);
            byte[] bResult = System.Text.Encoding.Convert(srcEnc, destEnc, bData);
            return destEnc.GetString(bResult);
        }

        /// <summary>
        /// fills route list with point names and road segment names
        /// </summary>
        /// <param name="tbl">path table (lat, lon, name)</param>
        private void FillRouteList(DeliveryCarResult carResult)
        {
            var last_street = "";
            foreach (var resultPoint in carResult.Route)
            {
                var ptype = resultPoint.Type;
                if (ptype == 10)
                {
                    var street = resultPoint.Description;
                    if (string.IsNullOrEmpty(street) || string.IsNullOrWhiteSpace(street)) continue; // street = "<no name>";
                    if (last_street == street) continue;
                    lb_route_descr.Items.Add(string.Format("{0} - {1}", resultPoint.ArrivalTime, street));
                    last_street = street;
                    continue;
                }
                var pointdescr = string.Format("{0} - {1}", resultPoint.ArrivalTime, point_types[ptype]);
                    
                if ((ptype == 1 || ptype == 2) && resultPoint.OrderId.HasValue) // 1-load, 2-unload
                    pointdescr += " job " + GetOrderNameByHashId(resultPoint.OrderId.Value);
                if (resultPoint.OrderId.HasValue)
                    pointdescr += " [" + GetOrderDescriptionByHashId(resultPoint.OrderId.Value, ptype == 1) + "]";

                lb_route_descr.Items.Add(pointdescr);
            }
        }

        private Point grid_jobs_last_point = new Point(-1, -1);
        private int grid_jobs_dragged_index = -1;
        private int grid_jobs_drop_min_index, grid_jobs_drop_max_index;
        private string dragJobName = "";
        private string dragJobCarName = "";
        private int jobNameColumn = 0;
        private string carOfJobList = ""; // name of car for which the resulting job list is shown
        List<GeoPoint> routeGeoPoints = new List<GeoPoint>(); // accordingly to the currently shown points in job_grid for one car or empty

        /// <summary>
        /// hit test job grid
        /// </summary>
        /// <param name="x">x</param>
        /// <param name="y">y</param>
        /// <param name="is_client_point">client point or screen point</param>
        /// <returns></returns>
        private int JobIndexFromPoint(int x, int y, bool is_client_point)
        {
            if (!is_client_point)
            {
                Point p = grid_jobs.PointToClient(new Point(x, y));
                x = p.X;
                y = p.Y;
            }
            DataGridView.HitTestInfo info = grid_jobs.HitTest(x, y);
            if (info.Type == DataGridViewHitTestType.Cell) return info.RowIndex;
            if (info.Type == DataGridViewHitTestType.None && grid_jobs.Rows.Count > 0)
            {
                Rectangle rect = grid_jobs.GetRowDisplayRectangle(grid_jobs.Rows.Count - 1, false);
                if (y > rect.Bottom && x < rect.Right)
                    return grid_jobs.Rows.Count; // after the last
            }
            return -1;
        }

        /// <summary>
        /// preparing to drag'n'drop
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_jobs_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button != MouseButtons.Left)
                return;
            int jobIx = JobIndexFromPoint(e.X, e.Y, true);

            grid_jobs_dragged_index = -1;
            grid_jobs_last_point.X = -1;
            grid_jobs_drop_min_index = grid_jobs_drop_max_index = 0;

            if (jobIx >= 0 && jobIx < grid_jobs.Rows.Count)
            {
                dragJobCarName = carOfJobList;
                dragJobName = grid_jobs.Rows[jobIx].Cells[jobNameColumn].Value.ToString(); // find out job name
                if (dragJobName != "")
                {
                    // prepare to drag
                    var job = _deliveryResult.GetOrderByName(dragJobName);
                    if (job != null)
                    {
                        dragJobCarName = GetCarNameByHashId(job.CarId);
                        grid_jobs_dragged_index = jobIx;
                        grid_jobs_last_point = e.Location;
                    }
                }
                else if (IsAuxPoint(jobIx))
                {// the dragged item is Lunch or Trip End
                    grid_jobs_dragged_index = jobIx;
                    grid_jobs_last_point = e.Location;
                }
            }
        }

        private bool IsAuxPoint(int ix)
        {
            if (jobNameColumn != 1 || ix < 0 || ix >= grid_jobs.Rows.Count)
                return false;
            string ptype = grid_jobs.Rows[ix].Cells[0].Value.ToString();
            return (ptype == point_types[4] || ptype == point_types[6]);
        }

        private bool DraggingAuxPoint()
        {
            return IsAuxPoint(grid_jobs_dragged_index);
        }

        /// <summary>
        /// are two points closer then delta?
        /// </summary>
        /// <param name="p1"></param>
        /// <param name="p2"></param>
        /// <param name="delta"></param>
        /// <returns></returns>
        private static bool InDelta(Point p1, Point p2, int delta)
        {
            int dx = p1.X - p2.X, dy = p1.Y - p2.Y;
            if ((Math.Abs(dx) > delta || Math.Abs(dy) > delta))
                return false;
            return dx * dx + dy * dy <= delta * delta;
        }

        /// <summary>
        /// if mouse moved enough begins drag and drop
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void grid_jobs_MouseMove(object sender, MouseEventArgs e)
        {
            const int delta = 3;
            Point p = e.Location;
            if (grid_jobs_dragged_index == -1 || grid_jobs_last_point.X == -1 || InDelta(grid_jobs_last_point, p, delta) ||
                e.Button != MouseButtons.Left)
                return;

            grid_jobs_last_point.X = -1;
            lastDropSource = grid_jobs;

            if (list_cars.Items[0].Selected || list_cars.Items[list_cars.Items.Count - 1].Selected)
                grid_jobs_drop_min_index = grid_jobs_drop_max_index = 0;
            // не имеет смысла переносить заказы в пределах маршрута, если они показаны все или нераспределенные
            else
            {
                grid_jobs_drop_min_index = grid_jobs_drop_max_index = grid_jobs_dragged_index;

                int minix = 0, maxix = grid_jobs.Rows.Count; // minimal and maximal possible drop index
                if (grid_jobs.Rows[minix].Cells[jobNameColumn].Value.ToString() == "") minix++; // begin point
                if (grid_jobs.Rows[maxix - 1].Cells[jobNameColumn].Value.ToString() == "") maxix--; // end point

                if (DraggingAuxPoint()) 
                {
                    // dont move lunch
                    grid_jobs_drop_min_index = grid_jobs_drop_max_index = 0;
                    //grid_jobs_drop_min_index = minix;
                    //grid_jobs_drop_max_index = maxix;
                }
                else
                {
                    for (int i = minix; i < maxix; i++)
                    {
                        // searching other inex of this job
                        if (i == grid_jobs_dragged_index) continue; // we need other index
                        string jname = grid_jobs.Rows[i].Cells[jobNameColumn].Value.ToString();
                        if (jname == dragJobName) // i is second index
                        {
                            if (i > grid_jobs_dragged_index) // dragging before i
                            {
                                grid_jobs_drop_min_index = minix;
                                grid_jobs_drop_max_index = i;
                            }
                            else // dragging after
                            {
                                grid_jobs_drop_min_index = i + 1;
                                grid_jobs_drop_max_index = maxix;
                            }
                            break;
                        }
                    }
                }
            }

            DragDropEffects dde = DoDragDrop(grid_jobs, DragDropEffects.Move);
            lastDropSource = null;
            grid_jobs_dragged_index = -1;
            InvalidateDropTarget();
        }

        private void grid_jobs_MouseUp(object sender, MouseEventArgs e)
        {
            grid_jobs_last_point.X = -1;
            grid_jobs_dragged_index = -1;
        }

        private void list_cars_DragEnter(object sender, DragEventArgs e)
        {
        }

        private Control lastDropTarget = null;
        private Control lastDropSource = null;
        private int lastDropTargetIndex = -1;
        private Rectangle lastDropTargetInvalidateRect = Rectangle.Empty;

        private void InvalidateDropTarget()
        {
            if (lastDropTarget != null && !lastDropTargetInvalidateRect.IsEmpty)
                lastDropTarget.Invalidate(lastDropTargetInvalidateRect);
            lastDropTarget = null;
            lastDropTargetIndex = -1;
            lastDropTargetInvalidateRect = Rectangle.Empty;
        }

        /// <summary>
        /// fills with semi-transparent red the area of drop target
        /// </summary>
        /// <param name="target"></param>
        /// <param name="index"></param>
        private void SetDropTarget(Control target, int index)
        {
            if (lastDropTarget == target && lastDropTargetIndex == index)
                return;
            InvalidateDropTarget();
            if (target == null)
                return;
            lastDropTarget = target;
            lastDropTargetIndex = index;

            if (target == list_cars)
            {
                lastDropTargetInvalidateRect = list_cars.Items[index].Bounds;
            }
            else if (target == grid_jobs)
            {
                if (index == grid_jobs.Rows.Count)
                {
                    lastDropTargetInvalidateRect = grid_jobs.GetRowDisplayRectangle(index - 1, true);
                    lastDropTargetInvalidateRect.Offset(0, lastDropTargetInvalidateRect.Height);
                }
                else lastDropTargetInvalidateRect = grid_jobs.GetRowDisplayRectangle(index, true);
                lastDropTargetInvalidateRect.Height = 8;
                lastDropTargetInvalidateRect.Offset(0, -4);
            }
            using (Graphics gr = target.CreateGraphics())
            using (Brush br = new SolidBrush(Color.FromArgb(100, Color.MediumVioletRed)))
                gr.FillRectangle(br, lastDropTargetInvalidateRect);
        }


        private void list_cars_DragLeave(object sender, EventArgs e)
        {
            SetDropTarget(null, -1);
        }

        /// <summary>
        /// hit test of car list
        /// </summary>
        /// <param name="e"></param>
        /// <returns></returns>
        private int CarIndexFromPoint(DragEventArgs e)
        {
            Point p = list_cars.PointToClient(new Point(e.X, e.Y));
            ListViewHitTestInfo hti = list_cars.HitTest(p);
            if (hti.Item == null)
                return -1;
            if (p.Y < hti.Item.Bounds.Top || p.Y >= hti.Item.Bounds.Bottom)
                return -1;
            if (hti.Item.Selected) // not for already selected car
                return -1;
            return hti.Item.Index;
        }

        private void list_cars_DragOver(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs || DraggingAuxPoint())
                return;
            int carIx = CarIndexFromPoint(e);
            if (carIx > 0 && // not -1 and not "all cars"
                dragJobCarName != list_cars.Items[carIx].Text) // not to the same car
            {
                e.Effect = DragDropEffects.Move;
                SetDropTarget(list_cars, carIx);
            }
            else
            {
                e.Effect = DragDropEffects.None;
                SetDropTarget(null, -1);
            }
        }

        private void grid_jobs_DragEnter(object sender, DragEventArgs e)
        {
        }

        private void grid_jobs_DragLeave(object sender, EventArgs e)
        {
            SetDropTarget(null, -1);
        }

        private void grid_jobs_DragOver(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs)
                return;
            // we do not move job if there are all connected jobs shown
            int jobIx = grid_jobs_drop_min_index != grid_jobs_drop_max_index ? JobIndexFromPoint(e.X, e.Y, false) : -1;

            if (jobIx >= grid_jobs_drop_min_index && jobIx <= grid_jobs_drop_max_index && // 0 is route beginning
                jobIx != grid_jobs_dragged_index && jobIx != grid_jobs_dragged_index + 1) // not adjacent
            {
                e.Effect = DragDropEffects.Move;
                SetDropTarget(grid_jobs, jobIx);
            }
            else
            {
                e.Effect = DragDropEffects.None;
                SetDropTarget(null, -1);
            }


        }

        private void list_cars_DragDrop(object sender, DragEventArgs e)
        {
            if (e.Effect != DragDropEffects.Move)
                return;
            int carIx = CarIndexFromPoint(e);
            if (carIx <= 0) // not to "all cars"
                return;
            string targetCarName = carIx < list_cars.Items.Count - 1 ? list_cars.Items[carIx].Text : "";
            if (dragJobCarName == targetCarName) // not to the same car
                return;

            MoveDraggedJobToCar(targetCarName);
        }

        private void grid_jobs_DragDrop(object sender, DragEventArgs e)
        {
            if (lastDropSource != grid_jobs)
                return;

            if (grid_jobs_drop_min_index == grid_jobs_drop_max_index)
                return;
            int jobIx = JobIndexFromPoint(e.X, e.Y, false);

            if (jobIx < grid_jobs_drop_min_index || jobIx > grid_jobs_drop_max_index || // overrun
                jobIx == grid_jobs_dragged_index || jobIx == grid_jobs_dragged_index + 1) // adjacent
                return;

            MoveDraggedJobToPosition(jobIx);
            grid_jobs.Rows[grid_jobs_dragged_index < jobIx ? jobIx - 1 : jobIx].Selected = true;
        }

        /// <summary>
        /// assigning new car to job (put dropped job in target car)
        /// </summary>
        /// <param name="targetCarName"></param>
        private void MoveDraggedJobToCar(string targetCarName)
        {
            WaitCursor wc = new WaitCursor();
            var job = _deliveryResult.GetOrderByName(dragJobName);
            var carId = string.IsNullOrEmpty(targetCarName) ? 0 : (uint)targetCarName.GetHashCode();
            var request = _deliveryRequest.MoveOrderToCar(carId, job.OrderId, _deliveryResult);
            var res = GetDelivery(request);
            if (string.IsNullOrEmpty(res))
            {
                UpdateCarAndJobLists(true);
                //MessageBox.Show("Empty result. Can not move job.");
                return;
            }
            try
            {
                var errorResult = JsonConvert.DeserializeObject<List<string>>(res);
                MessageBox.Show(string.Format("Error: {0}", res));
                return;
            }
            catch (Exception)
            {
            }
            var deliveryResult = JsonConvert.DeserializeObject<DeliveryResult>(res);
            _deliveryResult.UpdateResult(deliveryResult);
            UpdateCarAndJobLists(true);
        }

        /// <summary>
        /// assigning new point index in route in the same car (move dropped point in target position)
        /// </summary>
        /// <param name="jobIx"></param>
        private void MoveDraggedJobToPosition(int jobIx)
        {
            WaitCursor wc = new WaitCursor();

            if (DraggingAuxPoint()) // now do nothing if dragging lunch
            {
                //GWXDeliveryCar car = delivery.GetCar(dragJobCarName);
                //car.ChangeRoutePointIndex(grid_jobs_dragged_index, jobIx);
            }
            else
            {
                var job = _deliveryResult.GetOrderByName(dragJobName);
                var isLoadingPoint = IsLoadingPoint(dragJobName, grid_jobs_dragged_index);
                var nextJobName = grid_jobs.Rows[jobIx].Cells[jobNameColumn].Value.ToString();
                var nextJob = _deliveryResult.GetOrderByName(nextJobName);
                var isNextLoading = IsLoadingPoint(nextJobName, jobIx);
                var request = _deliveryRequest.MoveOrderInsideCar(job.CarId, job.OrderId, isLoadingPoint, nextJob.OrderId, isNextLoading, _deliveryResult);
                var res = GetDelivery(request);
                if (string.IsNullOrEmpty(res))
                {
                    MessageBox.Show("Empty result. Can not move job.");
                    return;
                }
                try
                {
                    var errorResult = JsonConvert.DeserializeObject<List<string>>(res);
                    MessageBox.Show(string.Format("Error: {0}", res));
                    return;
                }
                catch (Exception)
                {
                }
                var deliveryResult = JsonConvert.DeserializeObject<DeliveryResult>(res);
                _deliveryResult.UpdateResult(deliveryResult);
            }
            UpdateCarAndJobLists();
        }

        private bool IsLoadingPoint(string jobName, int indexInList)
        {
            var i = indexInList - 1;
            while (i >= 0)
            {
                if (jobName == grid_jobs.Rows[i].Cells[jobNameColumn].Value.ToString())
                    return false;
                i--;
            }
            return true;
        }

        private DeliveryRequest MoveJobInsideCar(uint carId, uint jobId)
        {
            var result = new DeliveryRequest();

            return null;
        }

        private void UpdateDeliveryResult(DeliveryResult updateData)
        {
            // update _deliveryresult
        }

        private void UpdateCarAndJobLists(bool clearCarsList = false)
        {
            // updating cars list (at place)
            FillCarsList(clearCarsList);
            // updateing job list
            FillJobList();
        }

        /// <summary>
        /// initializing form on start up
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Form1_Shown(object sender, EventArgs e)
        {
            _routeVisualManager = new RoutesVisualManager(_mapControl);
            if (settings.lastdata != "")
                button_last_data_Click(sender, e); // load last selected data
        }

        private void grid_jobs_MouseDoubleClick(object sender, MouseEventArgs e)
        {// double click on grid job - must show position of job point on map
            //if (e.Button != MouseButtons.Left)
            //    return;
            //int pointIx = JobIndexFromPoint(e.X, e.Y, true);
            //if (pointIx == -1 || pointIx >= routeGeoPoints.Count)
            //    return;
            //GeoPoint gp = routeGeoPoints[pointIx];

            //axGWControl1.DeleteDBF("mark");
            //axGWControl1.Table2Map("name='mark';descr='Marked Positions';structure='name text';", "m @502 'gwp.otl' 2", null);
            //string coord = gp.Lon.ToString("g") + " " + gp.Lat.ToString("g");
            //axGWControl1.ModifyTable("insert into [mark] set [name]='" + coord + "', metrics='" + "P E " + coord + ";" + "'", 0);

            //axGWControl1.SetGeoCenter(gp.Lat, gp.Lon);
            //axGWControl1.Refresh();
        }

        private string GetSessionId()
        {
            var httpWebRequest = (HttpWebRequest)WebRequest.Create(ConfigurationManager.AppSettings["GwsUrl"] + "api/new_session");
            //httpWebRequest.ContentType = "text/json";
            httpWebRequest.Method = "GET";

            //using (var streamWriter = new StreamWriter(httpWebRequest.GetRequestStream()))
            //{
            //    string json = "{\"user\":\"test\"," +
            //                  "\"password\":\"bla\"}";

            //    streamWriter.Write(json);
            //    streamWriter.Flush();
            //    streamWriter.Close();
            //}

            var httpResponse = (HttpWebResponse)httpWebRequest.GetResponse();
            using (var streamReader = new StreamReader(httpResponse.GetResponseStream()))
            {
                return streamReader.ReadToEnd();
            }
        }

        private string GetDelivery(DeliveryRequest deliveryRequest)
        {
            var httpWebRequest = (HttpWebRequest)WebRequest.Create(ConfigurationManager.AppSettings["GwsUrl"] + "api/delivery.gws");
            httpWebRequest.ContentType = "text/json";
            httpWebRequest.Method = "POST";
            httpWebRequest.Timeout = 60*5*1000;

            using (var streamWriter = new StreamWriter(httpWebRequest.GetRequestStream()))
            {
                var request = JsonConvert.SerializeObject(deliveryRequest);
                streamWriter.Write(request);
                streamWriter.Flush();
                streamWriter.Close();
            }

            var httpResponse = (HttpWebResponse)httpWebRequest.GetResponse();
            using (var streamReader = new StreamReader(httpResponse.GetResponseStream()))
            {
                return streamReader.ReadToEnd();
            }
        }

        private RoadVariants GetRoadVariants()
        {
            try
            {
                var httpWebRequest = (HttpWebRequest)WebRequest.Create(ConfigurationManager.AppSettings["GwsUrl"] + "api/road_variants?sid=" + _sessionId);
                var httpResponse = (HttpWebResponse)httpWebRequest.GetResponse();
                using (var streamReader = new StreamReader(httpResponse.GetResponseStream()))
                {
                    return JsonConvert.DeserializeObject<RoadVariants>(streamReader.ReadToEnd());
                }
            }
            catch (Exception)
            {
                return null;
            }
        }

        private void FormGwsApi_Load(object sender, System.EventArgs e)
        {
            _mapControl = new Map { Center = new Location(59.940770, 30.313238), ZoomLevel = 11, Focusable = true };
            _mapControl.MouseEnter += (o, args) => _mapControl.Focus();

            var tileLayer = new TileLayer();
            tileLayer.TileSource = new TileSource { UriFormat = ConfigurationManager.AppSettings["GwsUrl"] + "api/tile/{z}/{x}/{y}?sid=" + _sessionId };
            _mapControl.TileLayers.Add(tileLayer);

            _ctrlHost = new ElementHost { Dock = DockStyle.Fill };
            MapControlHost.Controls.Add(_ctrlHost);
            _ctrlHost.Child = _mapControl;

            SetDeliveryPropertiesControls();
            //toolStripStatusLabel1.Text = st;

            linkLabel1.Links.Remove(linkLabel1.Links[0]);
            linkLabel1.Links.Add(0, linkLabel1.Text.Length, "http://www.samlogist.com/");
            ShowDataOnMap();
        }

        private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            var sInfo = new ProcessStartInfo(e.Link.LinkData.ToString());
            Process.Start(sInfo);
        }
    }
}