﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Newtonsoft.Json;

namespace GWXDeliveryDemo
{
    public class GeoPoint
    {
        [JsonProperty("lat")] public double Latitude;
        [JsonProperty("lon")] public double Longtitude;
    }

    public class DeliveryParameters
    {
        [JsonProperty("optimizationMode")] public string OptimizationMode;
        [JsonProperty("distributionMethod")] public string DistributionMethod;
        [JsonProperty("multiTripTimeInterval")] public uint MultiTripTimeInterval;
        [JsonProperty("countPathFromGarage")] public bool CountPathFromGarage;
        [JsonProperty("countPathToGarage")] public bool CountPathToGarage;
        [JsonProperty("allowGarageInMultiTrip")] public bool AllowGarageInMultiTrip;
        [JsonProperty("visitEachPointOnlyOnce")] public bool VisitEachPointOnlyOnce;
        [JsonProperty("allowMultiTripRoutes")] public bool AllowMultiTripRoutes;
        [JsonProperty("useStackUnloading")] public bool UseStackUnloading;
        [JsonProperty("isGroupDelivery")] public bool IsGroupDelivery;
        [JsonProperty("includeAdditionsInRoute")] public bool IncludeAdditionsInRoute;
        [JsonProperty("additionalDistanceInPoint")] public int AdditionalDistanceInPoint;
        [JsonProperty("additionalTimeInPoint")] public int AdditionalTimeInPoint;
        [JsonProperty("multiDayRoutes")] public bool MultiDayRoutes;
        [JsonProperty("useTollRoads")] public bool UseTollRoads;
    }

    public class DeliveryCar
    {
        [JsonProperty("id")] public uint Id;
        [JsonProperty("initialPosition")] public GeoPoint InitialPosition;
        [JsonProperty("maxWeight")] public uint MaxWeight = 0;
        [JsonProperty("maxVolume")] public uint MaxVolume = 0;
        [JsonProperty("maxVolume2")] public uint MaxVolume2 = 0;
        [JsonProperty("maxCost")] public uint MaxCost = 0;
        [JsonProperty("speedCoefficient")] public double SpeedCoefficient = 1;
        [JsonProperty("selfWeight")] public uint SelfWeight = 0;
        [JsonProperty("maxRouteLength")] public uint MaxRouteLength = 0;
        [JsonProperty("maxRouteTime")] public uint MaxRouteTime = 0;
        [JsonProperty("fuelConsumptionPerKm")] public double FuelConsumptionPerKm = 0;
        [JsonProperty("workStartTime")] public string WorkStartTime = "";
        [JsonProperty("workFinishTime")] public string WorkFinishTime = "";
        [JsonProperty("lunchStartTime")] public string LunchStartTime = "";
        [JsonProperty("lunchFinishTime")] public string LunchFinishTime = "";
        [JsonProperty("maxOrdersCount")] public uint MaxOrdersCount = 0;
        [JsonProperty("permit")] public string Permit = "";
        [JsonProperty("carType")] public uint CarType = 0;
        [JsonProperty("cargoType")] public uint CargoType = 0;
        [JsonProperty("width")] public uint Width = 0;
        [JsonProperty("height")] public uint Height= 0;
        [JsonProperty("allowedLoadZones")] public List<uint> AllowedLoadZones = new List<uint>();
        [JsonProperty("allowedUnloadZones")] public List<uint> AllowedUnloadZones = new List<uint>();
        [JsonProperty("workTime")] public string WorkTime = "";
        [JsonProperty("useTollRoads")] public int UseTollRoads = -1;

        public void FillZoneListFromString(string zoneList, bool isLoadZones)
        {
            foreach (var zoneName in zoneList.Split(';').Where(zoneName => !string.IsNullOrEmpty(zoneName)))
            {
                if(isLoadZones)
                    AllowedLoadZones.Add((uint)zoneName.GetHashCode());
                else
                    AllowedUnloadZones.Add((uint)zoneName.GetHashCode());
            }
        }
    }

    public class DeliveryOrder
    {
        [JsonProperty("id")] public uint Id;
        [JsonProperty("loadPosition")] public GeoPoint LoadPosition;
        [JsonProperty("handOverPosition")] public GeoPoint HandOverPosition;
        [JsonProperty("loadStartTime")] public string LoadStartTime = "";
        [JsonProperty("loadFinishTime")] public string LoadFinishTime = "";
        [JsonProperty("handOverStartTime")] public string HandOverStartTime = "";
        [JsonProperty("handOverFinishTime")] public string HandOverFinishTime = "";
        [JsonProperty("priority")] public uint Priority = 0;
        [JsonProperty("loadTime")] public string LoadTime = "";
        [JsonProperty("handOverTime")] public string HandOverTime = "";
        [JsonProperty("carTypes")] public List<uint> CarTypes = new List<uint>();
        [JsonProperty("weight")] public uint Weight = 0;
        [JsonProperty("volume")] public uint Volume = 0;
        [JsonProperty("carVolumeUsage")] public uint CarVolumeUsage = 0;
        [JsonProperty("cost")] public uint Cost = 0;
        [JsonProperty("loadDuration")] public uint LoadDuration = 0;
        [JsonProperty("handOverDuration")] public uint HandOverDuration = 0;
        [JsonProperty("loadPointIndexInRoute")] public int LoadPointIndexInRoute = -1;
        [JsonProperty("handOverPointIndexInRoute")] public int HandOverPointIndexInRoute = -1;
        [JsonProperty("maxTransportationTime")] public uint MaxTransportationTime = 0;
        [JsonProperty("attachedCarId")] public uint AttachedCarId = 0;

        public void FillCarTypesFromString(string carTypes)
        {
            foreach (var carType in carTypes.Split(';').Where(carType => !string.IsNullOrEmpty(carType)))
            {
                CarTypes.Add((uint)carType.GetHashCode());
            }
        }

        public void SetPointIndex(int index, bool isLoading)
        {
            if (isLoading)
                LoadPointIndexInRoute = index;
            else
                HandOverPointIndexInRoute = index;
        }
    }

    public class DeliveryZone
    {
        [JsonProperty("id")] public uint Id;
        [JsonProperty("metrics")] public string Metrics;
    }

    public class DeliveryRequestData
    {
        [JsonProperty("parameters")] public DeliveryParameters Parameters = new DeliveryParameters();
        [JsonProperty("cars")] public List<DeliveryCar> Cars = new List<DeliveryCar>();
        [JsonProperty("orders")] public List<DeliveryOrder> Orders = new List<DeliveryOrder>();
        [JsonProperty("zones")] public List<DeliveryZone> Zones = new List<DeliveryZone>();
    }

    public class DeliveryRequest
    {
        [JsonProperty("sessionId")] public uint SessionId;
        [JsonProperty("accessKey")] public string AccessKey;
        [JsonProperty("deliveryData")] public DeliveryRequestData DeliveryData = new DeliveryRequestData();

        public DeliveryRequest MoveOrderInsideCar(uint carId, uint jobId, bool isLoadingPoint, uint nextJobId, bool isNextLoading, DeliveryResult deliveryResult)
        {
            var result = new DeliveryRequest
            {
                AccessKey = AccessKey,
                SessionId = SessionId,
                DeliveryData = {Parameters = DeliveryData.Parameters}
            };
            result.DeliveryData.Parameters.DistributionMethod = "preDistributed";
            var car = DeliveryData.Cars.Single(c => c.Id == carId);
            result.DeliveryData.Cars.Add(car);

            // add all orders
            foreach (var orderId in deliveryResult.GetAllOrdersByCarId(carId))
            {
                var order = DeliveryData.Orders.Single(o => o.Id == orderId);
                order.AttachedCarId = carId;
                result.DeliveryData.Orders.Add(order);
            }

            // set correct indices
            var pointIndex = 1;
            foreach (var point in deliveryResult.Cars.Single(c=>c.CarId==carId).Route.Where(p => p.Type == 1 || p.Type == 2))
            {
                var sameOrderAction = (isLoadingPoint && point.Type == 1) || (!isLoadingPoint && point.Type == 2);
                if (point.OrderId == jobId && sameOrderAction)
                {
                    continue;
                }

                sameOrderAction = (isNextLoading && point.Type == 1) || (!isNextLoading && point.Type == 2);
                if (point.OrderId == nextJobId && sameOrderAction)
                {
                    result.SetOrderPointIndex(jobId, pointIndex, isLoadingPoint);
                    pointIndex++;
                }

                if (point.OrderId == null) continue;
                
                result.SetOrderPointIndex(point.OrderId.Value, pointIndex, point.Type == 1);
                pointIndex++;
            }
            
            return result;
        }

        public DeliveryRequest MoveOrderToCar(uint carId, uint jobId, DeliveryResult deliveryResult)
        {
            var result = new DeliveryRequest
            {
                AccessKey = AccessKey,
                SessionId = SessionId,
                DeliveryData = { Parameters = DeliveryData.Parameters }
            };
            result.DeliveryData.Parameters.DistributionMethod = "preDistributed";

            var calculatedOrder = deliveryResult.Orders.Single(o => o.OrderId == jobId);
            
            // moved from
            if (calculatedOrder.CarId != 0)
            {
                var car = DeliveryData.Cars.Single(c => c.Id == calculatedOrder.CarId);
                result.DeliveryData.Cars.Add(car);

                foreach (var orderId in deliveryResult.GetAllOrdersByCarId(calculatedOrder.CarId))
                {
                    if (orderId == jobId) continue;
                    
                    var order = DeliveryData.Orders.Single(o => o.Id == orderId);
                    order.AttachedCarId = calculatedOrder.CarId;
                    order.LoadPointIndexInRoute = -1;
                    order.HandOverPointIndexInRoute = -1;
                    result.DeliveryData.Orders.Add(order);
                }

                deliveryResult.ClearCarResult(calculatedOrder.CarId); 
                
                calculatedOrder.CarId = 0;
                calculatedOrder.LoadTime = "";
                calculatedOrder.HandOverTime = "";
            }

            // move to
            if (carId != 0)
            {
                var car = DeliveryData.Cars.Single(c => c.Id == carId);
                result.DeliveryData.Cars.Add(car);

                foreach (var orderId in deliveryResult.GetAllOrdersByCarId(carId))
                {
                    var order = DeliveryData.Orders.Single(o => o.Id == orderId);
                    order.AttachedCarId = carId;
                    order.LoadPointIndexInRoute = -1;
                    order.HandOverPointIndexInRoute = -1;
                    result.DeliveryData.Orders.Add(order);
                }

                var movingOrder = DeliveryData.Orders.Single(o => o.Id == jobId);
                movingOrder.AttachedCarId = carId;
                movingOrder.LoadPointIndexInRoute = -1;
                movingOrder.HandOverPointIndexInRoute = -1;
                result.DeliveryData.Orders.Add(movingOrder);

                deliveryResult.ClearCarResult(carId); 
            }
            
            return result;
        }

        public void SetOrderPointIndex(uint orderId, int index, bool isLoading)
        {
            DeliveryData.Orders.Single(o => o.Id == orderId).SetPointIndex(index, isLoading);
        }
    }

    public class DeliveryResultPoint
    {
        [JsonProperty("coordinates")] public GeoPoint Coordinates;
        [JsonProperty("description")] public string Description;
        [JsonProperty("arrivalTime")] public DateTime ArrivalTime;
        [JsonProperty("type")] public uint Type;
        [JsonProperty("orderId")] public uint? OrderId;
        [JsonProperty("volumeUsed")] public uint? VolumeUsed;
        [JsonProperty("errorMessage")] public string ErrorMessage;
    }

    public class DeliveryOrderResult
    {
        [JsonProperty("orderId")] public uint OrderId;
        [JsonProperty("carId")] public uint CarId;
        [JsonProperty("loadTime")] public string LoadTime;
        [JsonProperty("handOverTime")] public string HandOverTime;
    }

    public class DeliveryCarResult
    {
        [JsonProperty("carId")] public uint CarId;
        [JsonProperty("totalRouteLength")] public uint TotalRouteLength;
        [JsonProperty("totalRouteTime")] public uint TotalRouteTime;
        [JsonProperty("totalOrdersCount")] public uint TotalOrdersCount;
        [JsonProperty("routeStartTime")] public string RouteStartTime;
        [JsonProperty("routeEndTime")] public string RouteEndTime;
        [JsonProperty("errorPointsCount")] public uint ErrorPointsCount;
        [JsonProperty("route")] public List<DeliveryResultPoint> Route;
    }

    public class DeliveryResult
    {
        [JsonProperty("cars")] public List<DeliveryCarResult> Cars;
        [JsonProperty("orders")] public List<DeliveryOrderResult> Orders;

        public DeliveryOrderResult GetOrderByName(string name)
        {
            return Orders.FirstOrDefault(order => order.OrderId == (uint) name.GetHashCode());
        }

        public List<uint> GetAllOrdersByCarId(uint carId)
        {
            return Orders.Where(o => o.CarId == carId).Select(o => o.OrderId).ToList();
        }

        public int GetOrderIndex(uint orderId, bool isLoading)
        {
            var result = -1;
            var car = Cars.Single(c => c.CarId == Orders.Single(o => o.OrderId == orderId).CarId);
            foreach (var point in car.Route.Where(point => point.Type != 10))
            {
                result++;
                if (point.OrderId == orderId && point.Type == 1 && isLoading)
                    return result;
                if (point.OrderId == orderId && point.Type == 2 && !isLoading)
                    return result;
            }
            return result;
        }

        public void UpdateResult(DeliveryResult newData)
        {
            foreach (var car in newData.Cars)
            {
                Cars.RemoveAll(c => c.CarId == car.CarId);
                Cars.Add(car);
            }

            foreach (var order in newData.Orders)
            {
                Orders.RemoveAll(o => o.OrderId == order.OrderId);
                Orders.Add(order);
            }
        }

        public void ClearCarResult(uint carId)
        {
            var car = Cars.Single(c => c.CarId == carId);
            car.TotalOrdersCount = 0;
            car.TotalRouteLength = 0;
            car.TotalRouteTime = 0;
            car.Route.Clear();
            car.ErrorPointsCount = 0;
            car.RouteEndTime = "";
            car.RouteStartTime = "";
        }
    }

    public class RoadVariants
    {
        [JsonProperty("width")] public List<uint> AvailableWidth;
        [JsonProperty("height")] public List<uint> AvailableHeight;
        [JsonProperty("weight")] public List<uint> AvailableWeight;
        [JsonProperty("priority")] public List<uint> AvailablePriority;
        [JsonProperty("cargo")] public List<double> AvailableCargo;
        [JsonProperty("permit")] public List<string> AvailablePermit;
    }
}