using System;
using System.Collections.Generic;
using System.Text;
using System.Data;

namespace DeliveryDemo
{
	class MyReader
	{
		private IDataReader _orig;
		public MyReader(IDataReader orig)
		{
			_orig = orig;
		}
		private int ToIndex(object ident)
		{
			if (ident is string)
				return GetOrdinal(ident.ToString());
			return Convert.ToInt32(ident);
		}
		public bool Read()
		{
			return _orig.Read();
		}
		public bool IsDBNull(object ident)
		{
			int ordinal = ToIndex(ident);
			if (ordinal < 0)
				return true;
			return _orig.IsDBNull(ordinal);
		}
		public string GetString(object ident)
		{
			int ordinal = ToIndex(ident);
			if (IsDBNull(ordinal))
				return "";
			return _orig.GetString(ordinal);
		}
		public double GetDouble(object ident)
		{
			return GetDouble(ident, 0);
		}
		public double GetDouble(object ident, double defval)
		{
			int ordinal = ToIndex(ident);
			if (IsDBNull(ordinal))
				return defval;
            try
            {
                return _orig.GetDouble(ordinal);
            }
            catch (Exception)
            {
            }
            try
            {
                object val = _orig.GetValue(ordinal);
                string str = val.ToString();//  _orig.GetString(ordinal);
                str = str.Replace(',', '.');
                return double.Parse(str, System.Globalization.CultureInfo.InvariantCulture);
            }
            catch (Exception)
            {
            }
            return 0.0;
		}
		public int GetInt32(object ident, int defaultValue = 0)
		{
			int ordinal = ToIndex(ident);
			if (IsDBNull(ordinal))
                return defaultValue;
			return _orig.GetInt32(ordinal);
		}
		public DateTime GetDateTime(object ident)
		{
			DateTime res;
			if (GetDateTime(ident, out res))
				return res;
			return new DateTime(0);
		}
		public bool GetDateTime(object ident, out DateTime res)
		{
			int ordinal = ToIndex(ident);
			if (IsDBNull(ordinal))
			{
				res = new DateTime(0);
				return false;
			}
			res = _orig.GetDateTime(ordinal);
			return true;
		}
		public int GetOrdinal(string name)
		{
			try
			{
				return _orig.GetOrdinal(name);
			}
			catch (IndexOutOfRangeException)
			{
				return -1;
			}
			catch (ArgumentException)
			{
				return -1;
			}
			catch (Exception)
			{
				return -1;
			}

		}
	}
}
