﻿using System.Linq;
using MapControl;
using System.Collections.Generic;

namespace GWXDeliveryDemo
{
    public class RouteVisualInfo
    {
        public int Id;

        readonly Map _mapControl;
        readonly List<RouteSegmentVisualInfo> _segments;
        readonly List<RouteStopVisualInfo> _stops;
        public RouteVisualInfo(Map mapControl)
        {
            _mapControl = mapControl;
            _segments = new List<RouteSegmentVisualInfo>();
            _stops = new List<RouteStopVisualInfo>();
        }

        public void AddSegment(IEnumerable<Location> points, string text)
        {
            _segments.Add(new RouteSegmentVisualInfo(_mapControl, points, text));
        }

        public void AddStop(Location point, string text, string shortText)
        {
            foreach (var stop in _stops.Where(stop => stop.IsSamePoint(point)))
            {
                stop.AddTooltipText(shortText);
                return;
            }

            _stops.Add(new RouteStopVisualInfo(_mapControl, point, text));
        }

        public void Clear()
        {
            foreach (var segment in _segments)
            {
                segment.Clear();
            }
            
            foreach (var stop in _stops)
            {
                stop.Clear();                
            }
            
            _segments.Clear();
            _stops.Clear();
        }
    }
}